---------------------------------------------------------------------------
-- Model of the 64-bit MIPS ISA (MIPS III with some extra instructions)
-- (c) Anthony Fox, University of Cambridge
---------------------------------------------------------------------------

--------------------------------------------------
-- Memory access
--------------------------------------------------

construct LorS  { LOAD, STORE }

bits(3) BYTE       = 0`3
bits(3) HALFWORD   = 1`3
bits(3) WORD       = 3`3
bits(3) DOUBLEWORD = 7`3

bool BigEndianMem = CP0.Config.BE
bits(1) ReverseEndian = [CP0.Status.RE and UserMode]
bits(1) BigEndianCPU  = [BigEndianMem] ?? ReverseEndian

type pAddr = bits(64)

declare MEM :: pAddr -> byte -- memory

pAddr * CCA AddressTranslation (vAddr::vAddr, LorS::LorS) =
   (vAddr, 2) -- null address translation

bool Aligned (vAddr::vAddr, MemType::bits(3)) = [vAddr] && MemType == 0

pAddr AdjustEndian (MemType::bits(3), pAddr::pAddr) =
  match MemType
  {
     case 0 => pAddr ?? [ReverseEndian^3]
     case 1 => pAddr ?? [ReverseEndian^2 : '0']
     case 3 => pAddr ?? [ReverseEndian : '00']
     case 7 => pAddr
     case _ => #UNPREDICTABLE ("bad access length")
  }

--------------------------------------------------
-- Loads
--------------------------------------------------

dword ReadData (a::pAddr) =
{
   a = a && ~0b111;
   if BigEndianMem then
     MEM(a)     : MEM(a + 1) : MEM(a + 2) : MEM(a + 3) :
     MEM(a + 4) : MEM(a + 5) : MEM(a + 6) : MEM(a + 7)
   else
     MEM(a + 7) : MEM(a + 6) : MEM(a + 5) : MEM(a + 4) :
     MEM(a + 3) : MEM(a + 2) : MEM(a + 1) : MEM(a)
}

dword LoadMemory (MemType::bits(3), AccessLength::bits(3), needAlign::bool,
                  vAddr::vAddr, link::bool option) =
  if needAlign and not Aligned (vAddr, MemType) then
  {
    CP0.BadVAddr <- vAddr;
    SignalException (AdEL);
    UNKNOWN
  }
  else
  {
    pAddr, _ = AddressTranslation (vAddr, LOAD);
    if exceptionSignalled then UNKNOWN
    else
    {
      pAddr = AdjustEndian (MemType, pAddr);
      match link
      {
        case Some (true) =>
          {
            LLbit <- Some (true);
            CP0.LLAddr <- [pAddr]
          }
        case Some (false) => LLbit <- None
        case None => nothing
      };
      ReadData (pAddr)
    }
  }

unit loadByte (base::reg, rt::reg, offset::bits(16), unsigned::bool) =
{
   vAddr = SignExtend (offset) + GPR(base);
   memdoubleword = LoadMemory (BYTE, BYTE, false, vAddr, Some (false));
   when not exceptionSignalled do
   {
      byte = vAddr<2:0> ?? BigEndianCPU^3;
      membyte`8 = memdoubleword <7 + 8 * [byte] : 8 * [byte]>;
      GPR(rt) <- if unsigned then ZeroExtend (membyte) else SignExtend (membyte)
   }
}

unit loadHalf (base::reg, rt::reg, offset::bits(16), unsigned::bool) =
{
   vAddr = SignExtend (offset) + GPR(base);
   memdoubleword = LoadMemory (HALFWORD, HALFWORD, true, vAddr, Some (false));
   when not exceptionSignalled do
   {
     byte = vAddr<2:0> ?? (BigEndianCPU^2 : '0');
     memhalf`16 = memdoubleword <15 + 8 * [byte] : 8 * [byte]>;
     GPR(rt) <- if unsigned then ZeroExtend (memhalf) else SignExtend (memhalf)
   }
}

unit loadWord (link::bool, base::reg, rt::reg, offset::bits(16),
               unsigned::bool) =
{
   vAddr = SignExtend (offset) + GPR(base);
   memdoubleword = LoadMemory (WORD, WORD, true, vAddr, Some (link));
   when not exceptionSignalled do
   {
     byte = vAddr<2:0> ?? (BigEndianCPU : '00');
     memword`32 = memdoubleword <31 + 8 * [byte] : 8 * [byte]>;
     GPR(rt) <- if unsigned then ZeroExtend (memword) else SignExtend (memword)
   }
}

unit loadDoubleword (link::bool, base::reg, rt::reg, offset::bits(16)) =
{
   vAddr = SignExtend (offset) + GPR(base);
   memdoubleword =
     LoadMemory (DOUBLEWORD, DOUBLEWORD, true, vAddr, Some (link));
   when not exceptionSignalled do GPR(rt) <- memdoubleword
}

-- Instruction Fetch
word option Fetch =
{
   vAddr = PC;
   memdoubleword = LoadMemory (WORD, WORD, true, vAddr, None);
   if exceptionSignalled then None
   else
   {
      bytesel = vAddr<2:0> ?? (BigEndianCPU : '00'); -- must be aligned
      memword = memdoubleword <31 + 8 * [bytesel] : 8 * [bytesel]>;
      Some (memword)
   }
}

--------------------------------------------------
-- Stores
--------------------------------------------------

unit WriteData (a::pAddr, MemElem::dword, l::nat, h::nat) =
{
   a = a && ~0b111;
   if BigEndianMem then
   {
     when l <= 7 and 7 <= h do MEM(a)     <- MemElem<63:56>;
     when l <= 6 and 6 <= h do MEM(a + 1) <- MemElem<55:48>;
     when l <= 5 and 5 <= h do MEM(a + 2) <- MemElem<47:40>;
     when l <= 4 and 4 <= h do MEM(a + 3) <- MemElem<39:32>;
     when l <= 3 and 3 <= h do MEM(a + 4) <- MemElem<31:24>;
     when l <= 2 and 2 <= h do MEM(a + 5) <- MemElem<23:16>;
     when l <= 1 and 1 <= h do MEM(a + 6) <- MemElem<15:8>;
     when l == 0            do MEM(a + 7) <- MemElem<7:0>
   }
   else
   {
     when l <= 7 and 7 <= h do MEM(a)     <- MemElem<7:0>;
     when l <= 6 and 6 <= h do MEM(a + 1) <- MemElem<15:8>;
     when l <= 5 and 5 <= h do MEM(a + 2) <- MemElem<23:16>;
     when l <= 4 and 4 <= h do MEM(a + 3) <- MemElem<31:24>;
     when l <= 3 and 3 <= h do MEM(a + 4) <- MemElem<39:32>;
     when l <= 2 and 2 <= h do MEM(a + 5) <- MemElem<47:40>;
     when l <= 1 and 1 <= h do MEM(a + 6) <- MemElem<55:48>;
     when l == 0            do MEM(a + 7) <- MemElem<63:56>
   }
}

bool StoreMemory (MemType::bits(3), AccessLength::bits(3), needAlign::bool,
                  MemElem::dword, vAddr::vAddr, cond::bool) =
  if needAlign and not Aligned (vAddr, MemType) then
  {
    CP0.BadVAddr <- vAddr;
    SignalException (AdES);
    false
  }
  else
  {
    pAddr, _ = AddressTranslation (vAddr, STORE);
    if exceptionSignalled then true
    else
    {
      pAddr = AdjustEndian (MemType, pAddr);
      sc_success = not cond or
        match LLbit
        {
          case None => #UNPREDICTABLE ("conditional store: LLbit not set")
          case Some (false) => false
          case Some (true) =>
            CP0.LLAddr == [pAddr] or
            #UNPREDICTABLE
               ("conditional store: address doesn't match previous LL address")
        };
      when sc_success do
      {
        b = [AccessLength] + 0n1;
        l = 8 - (b + [vAddr<2:0>]);
        h = l + b - 1;
        WriteData (pAddr, MemElem, l, h)
      };
      LLbit <- None;
      sc_success
    }
  }

bool storeWord (base::reg, rt::reg, offset::bits(16), cond::bool) =
{
   vAddr = SignExtend (offset) + GPR(base);
   bytesel = vAddr<2:0> ?? (BigEndianCPU : '00');
   datadoubleword = GPR(rt) << (0n8 * [bytesel]);
   StoreMemory (WORD, WORD, true, datadoubleword, vAddr, cond)
}

bool storeDoubleword (base::reg, rt::reg, offset::bits(16), cond::bool) =
{
   vAddr = SignExtend (offset) + GPR(base);
   datadoubleword = GPR(rt);
   StoreMemory (DOUBLEWORD, DOUBLEWORD, true, datadoubleword, vAddr, cond)
}
