-------------------------------------------------
-- Specification of x86-64                     --
-- (c) Anthony Fox, University of Cambridge    --
-------------------------------------------------

type byte = bits(8)
type word = bits(16)
type dword = bits(32)
type qword = bits(64)
type dqword = bits(128)

type stream = byte list
type ostream = stream option

---------------------------------------------------------------------------

-- ===============
-- The State Space
-- ===============

exception FAILURE :: string
exception BadFlagAccess :: string
exception INTERRUPT_EXCEPTION :: byte

unit DE_exception = #INTERRUPT_EXCEPTION (0)  -- Divide Error
unit UD_exception = #INTERRUPT_EXCEPTION (6)  -- Invalid Opcode
unit GP_exception = #INTERRUPT_EXCEPTION (13) -- General Protection
-- unit AC_exception = #INTERRUPT_EXCEPTION (17) -- Alignment Check
unit XM_exception = #INTERRUPT_EXCEPTION (19) -- SIMD Floating-point exception

------------
-- Registers
------------

-- General purpose registers

construct Zreg
  { RAX RCX RDX RBX RSP RBP RSI RDI zR8 zR9 zR10 zR11 zR12 zR13 zR14 zR15 }

type reg = Zreg -> qword

declare REG :: reg

-- The program counter

declare RIP :: qword

-- SSE registers

type xreg = bits(3)

register MXCSR :: dword {
  31-16  : Reserved
  15     : FZ         -- Flush to Zero
  14-13  : RC         -- Rounding Control
  12     : PM         -- Precision Mask
  11     : UM         -- Underflow Mask
  10     : OM         -- Overflow Mask
  9      : ZM         -- Divide-by-Zero Mask
  8      : DM         -- Denormal Operation Mask
  7      : IM         -- Invalid Operation Mask
  6      : DAZ        -- Denormals are Zeros
  5      : PE         -- Precision Flag
  4      : UE         -- Underflow Flag
  3      : OE         -- Overflow Flag
  2      : ZE         -- Divide-by-Zero Flag
  1      : DE         -- Denormal Operation Flag
  0      : IE         -- Invalid Operation Flag
}

declare {
  -- MMX_REG :: xreg -> qword
  XMM_REG :: xreg -> dqword
  MXCSR   :: MXCSR
}

---------
-- Memory
---------

type mem = qword -> byte

declare MEM :: mem

component mem8 (addr :: qword) :: byte
{
   value = MEM (addr)
   assign b = MEM (addr) <- b
}

component mem16 (addr :: qword) :: word
{
   value = ( mem8 (addr + 1) : mem8 (addr) )
   assign w =
      { mem8 (addr)     <- w<7:0>;
        mem8 (addr + 1) <- w<15:8>
      }
}

component mem32 (addr :: qword) :: dword
{
   value = ( mem16 (addr + 2) : mem16 (addr) )
   assign w =
      { mem16 (addr)     <- w<15:0>;
        mem16 (addr + 2) <- w<31:16>
      }
}

component mem64 (addr :: qword) :: qword
{
   value = ( mem32 (addr + 4) : mem32 (addr) )
   assign w =
      { mem32 (addr)     <- w<31:0>;
        mem32 (addr + 4) <- w<63:32>
      }
}

component mem128 (addr :: qword) :: dqword
{
   value = ( mem64 (addr + 8) : mem64 (addr) )
   assign w =
      { mem64 (addr)     <- w<63:0>;
        mem64 (addr + 8) <- w<127:64>
      }
}

--------
-- Flags
--------

construct Zeflags { Z_CF Z_PF Z_AF Z_ZF Z_SF Z_OF }

type eflags = Zeflags -> bool option

declare EFLAGS :: eflags

component Eflag (flag :: Zeflags) :: bool
{
   value =
      match EFLAGS (flag)
      {
         case Some (b) => b
         case None => #BadFlagAccess([flag])
      }
  assign b = EFLAGS (flag) <- Some (b)
}

unit FlagUnspecified (flag :: Zeflags) = EFLAGS (flag) <- None

component CF :: bool { value = Eflag (Z_CF) assign b = Eflag (Z_CF) <- b }
component PF :: bool { value = Eflag (Z_PF) assign b = Eflag (Z_PF) <- b }
component AF :: bool { value = Eflag (Z_AF) assign b = Eflag (Z_AF) <- b }
component ZF :: bool { value = Eflag (Z_ZF) assign b = Eflag (Z_ZF) <- b }
component SF :: bool { value = Eflag (Z_SF) assign b = Eflag (Z_SF) <- b }
component OF :: bool { value = Eflag (Z_OF) assign b = Eflag (Z_OF) <- b }

---------------------------------------------------------------------------

-- =========
-- AST types
-- =========

construct Zsize { Z8 :: bool Z16 Z32 Z64 }

construct Zbase
{
   ZnoBase
   ZripBase
   ZregBase :: Zreg
}

type maddr =  (bits(2) * Zreg) option * Zbase * qword

construct Zrm
{
   Zr :: Zreg  -- register
   Zm :: maddr -- mem [2^{scale} * index + base + displacement]
}

-- Here XX is one of 8, 16, 32, 64.
construct Zdest_src
{
   Zrm_i :: Zrm * qword   -- mnemonic r/mXX, immXX (sign-extended)
   Zrm_r :: Zrm * Zreg    -- mnemonic r/mXX, rXX
   Zr_rm :: Zreg * Zrm    -- mnemonic rXX, r/mXX
}

construct Zimm_rm
{
   Zrm  :: Zrm    -- r/mXX
   Zimm :: qword  -- sign-extended immediate
}

construct Zmonop_name { Zdec Zinc Znot Zneg }

construct Zbinop_name
{  Zadd Zor  Zadc Zsbb Zand Zsub Zxor  Zcmp
   Zrol Zror Zrcl Zrcr Zshl Zshr Ztest Zsar
}

construct Zbit_test_name { Zbt Zbts Zbtr Zbtc }

construct Zcond
{                 -- N = not
  Z_O    Z_NO     -- O = overflow
  Z_B    Z_NB     -- B = below
  Z_E    Z_NE     -- E = equal
  Z_NA   Z_A      -- A = above
  Z_S    Z_NS     -- S = signed
  Z_P    Z_NP     -- P = parity
  Z_L    Z_NL     -- L = less
  Z_NG   Z_G      -- L = greater
  Z_ALWAYS
}

---------------------------------------------------------------------------

-- =====================
-- Instruction Semantics
-- =====================

-- Effective addresses

construct Zea
{
   Zea_i :: Zsize * qword   -- Constant
   Zea_r :: Zsize * Zreg    -- Register name
   Zea_m :: Zsize * qword   -- Memory address
}

qword ea_index (index :: (bits(2) * Zreg) option) =
   match index
   {
      case None => 0
      case Some (scale, idx) => 1 << [scale] * REG (idx)
   }

qword ea_base (base :: Zbase) =
   match base
   {
      case ZnoBase => 0
      case ZripBase => RIP
      case ZregBase (b) => REG (b)
   }

qword mem_addr (m::maddr) =
{
  (index, base, displacement) = m;
  ea_index (index) + ea_base (base) + displacement
}

Zea ea_Zrm (size :: Zsize, rm :: Zrm) =
   match rm
   {
     case Zr (r) => Zea_r (size, r)
     case Zm (m) => Zea_m (size, mem_addr (m))
   }

Zea ea_Zdest (size :: Zsize, ds :: Zdest_src) =
   match ds
   {
      case Zrm_i (rm, _) => ea_Zrm (size, rm)
      case Zrm_r (rm, _) => ea_Zrm (size, rm)
      case Zr_rm (r, _)  => Zea_r (size, r)
   }

Zea ea_Zsrc (size :: Zsize, ds :: Zdest_src) =
   match ds
   {
      case Zrm_i (_, i)  => Zea_i (size, i)
      case Zrm_r (_, r)  => Zea_r (size, r)
      case Zr_rm (_, rm) => ea_Zrm (size, rm)
   }

Zea ea_Zimm_rm (imm_rm :: Zimm_rm) =
   match imm_rm
   {
      case Zrm (rm)   => ea_Zrm (Z64, rm)
      case Zimm (imm) => Zea_i (Z64, imm)
   }

-- Reading / Writing an EA

qword modSize (size :: Zsize, imm :: qword) =
   match size
   {
      case Z8 (_) => imm mod 8
      case Z16    => imm mod 16
      case Z32    => imm mod 32
      case Z64    => imm mod 64
   }

qword restrictSize (size :: Zsize, imm :: qword) =
   match size
   {
      case Z8 (_) => imm && 0xFF
      case Z16    => imm && 0xFFFF
      case Z32    => imm && 0xFFFFFFFF
      case Z64    => imm
   }

component EA (ea :: Zea) :: qword
{
   value =
      match ea
      {
         case Zea_i (i) => restrictSize (i)
         case Zea_r (Z8 (have_rex), r) =>
           (if have_rex or r notin set { RSP, RBP, RSI, RDI }
               then REG (r)
            else REG ([[r] - 0n4]) >>+ 8) && 0xFF
         case Zea_r (s, r)      => restrictSize (s, REG (r))
         case Zea_m (Z8 (_), a) => [mem8 (a)]
         case Zea_m (Z16, a)    => [mem16 (a)]
         case Zea_m (Z32, a)    => [mem32 (a)]
         case Zea_m (Z64, a)    => mem64 (a)
      }
   assign w =
      match ea
      {
         case Zea_i (i) => #FAILURE ("write to constant")
         case Zea_r (Z8 (have_rex), r) =>
            if have_rex or r notin set { RSP, RBP, RSI, RDI }
               then REG (r)<7:0> <- w<7:0>
            else REG ([[r] - 0n4])<15:8> <- w<7:0>
         case Zea_r (Z16, r)       => REG (r)<15:0> <- w<15:0>
         case Zea_r (Z32, r)       => REG (r) <- ZeroExtend (w<31:0>)
         case Zea_r (Z64, r)       => REG (r) <- w
         case Zea_m (Z8 (_), a)    => mem8 (a) <- w<7:0>
         case Zea_m (Z16, a)       => mem16 (a) <- w<15:0>
         case Zea_m (Z32, a)       => mem32 (a) <- w<31:0>
         case Zea_m (Z64, a)       => mem64 (a) <- w
      }
}

Zea * qword * qword read_dest_src_ea (sd :: Zsize * Zdest_src) =
{
   ea = ea_Zdest (sd);
   (ea, EA (ea), EA (ea_Zsrc (sd)))
}

-- Find the destination according to procedure call semantics
qword call_dest_from_ea (ea :: Zea) =
   match ea
   {
      case Zea_i (_, i) => RIP + i
      case Zea_r (_, r) => REG (r)
      case Zea_m (_, a) => mem64 (a)
   }

qword get_ea_address (ea :: Zea) =
   match ea
   {
      case Zea_i (_, i) => 0
      case Zea_r (_, r) => 0
      case Zea_m (_, a) => a
   }

-- RIP update

-- Update RIP according to procedure call
unit jump_to_ea (ea :: Zea) = RIP <- call_dest_from_ea (ea)

-- EFLAG updates

bool ByteParity (b :: byte) =
{  count = [b<7>] + [b<6>] + [b<5>] + [b<4>] +
           [b<3>] + [b<2>] + [b<1>] + [b<0>] :: nat;
   count mod 2 == 0
}

nat Zsize_width (size :: Zsize) =
   match size
   {
      case Z8 (_) => 8
      case Z16    => 16
      case Z32    => 32
      case Z64    => 64
   }

bool word_size_msb (size :: Zsize, w :: qword) = w<Zsize_width(size)-1>

unit write_PF (w :: qword) = PF <- ByteParity (w<7:0>)

unit write_SF (s_w :: Zsize * qword) = SF <- word_size_msb (s_w)

unit write_ZF (size :: Zsize, w :: qword) =
   ZF <- match size
         {
            case Z8 (_)  => [w] == 0`8
            case Z16     => [w] == 0`16
            case Z32     => [w] == 0`32
            case Z64     => w == 0
         }

unit write_arith_eflags_except_CF_OF (size :: Zsize, w :: qword) =
{
   write_PF (w);
   write_SF (size, w);
   write_ZF (size, w);
   FlagUnspecified (Z_AF)
}

unit write_arith_eflags (size :: Zsize, w :: qword, c :: bool, x :: bool) =
{
   CF <- c;
   OF <- x;
   write_arith_eflags_except_CF_OF (size, w)
}

unit write_logical_eflags (size :: Zsize, w :: qword) =
  write_arith_eflags (size, w, false, false)

unit erase_eflags = EFLAGS <- InitMap (None)

-- Bin-ops

nat value_width (s :: Zsize) = 2 ** Zsize_width (s)

bool word_signed_overflow_add (size :: Zsize, a :: qword, b :: qword) =
  ( word_size_msb (size, a) == word_size_msb (size, b) and
    word_size_msb (size, a + b) != word_size_msb (size, a) )

bool word_signed_overflow_sub (size :: Zsize, a :: qword, b :: qword) =
  ( word_size_msb (size, a) != word_size_msb (size, b) and
    word_size_msb (size, a - b) != word_size_msb (size, a) )

type result = qword * bool * bool

result add_with_carry_out (size :: Zsize, x :: qword, y :: qword) =
   (x + y, value_width (size) <= [x] + [y] :: nat,
    word_signed_overflow_add (size, x, y))

result sub_with_borrow (size :: Zsize, x :: qword, y :: qword) =
   (x - y, x <+ y, word_signed_overflow_sub (size, x, y))

unit write_arith_result (size :: Zsize, r :: result, ea :: Zea) =
{
   write_arith_eflags (size, r);
   EA (ea) <- Fst (r)
}

unit write_arith_result_no_CF_OF (size :: Zsize, w :: qword, ea :: Zea) =
{
   write_arith_eflags_except_CF_OF (size, w);
   EA (ea) <- w
}

unit write_logical_result (size :: Zsize, w :: qword, ea :: Zea) =
{
   write_logical_eflags (size, w);
   EA (ea) <- w
}

unit write_result_erase_eflags (w :: qword, ea :: Zea) =
{
   erase_eflags;
   EA (ea) <- w
}

qword SignExtension (w :: qword, size1 :: Zsize,  size2 :: Zsize) =
{
   var v = w;
   match size1, size2
   {
      case Z8 (_), Z16 => v<15:0> <- SignExtend (w<7:0>)
      case Z8 (_), Z32 => v<31:0> <- SignExtend (w<7:0>)
      case Z8 (_), Z64 => v <- SignExtend (w<7:0>)
      case Z16, Z32 => v<31:0> <- SignExtend (w<15:0>)
      case Z16, Z64 => v <- SignExtend (w<15:0>)
      case Z32, Z64 => v <- SignExtend (w<31:0>)
      case _ => #FAILURE "SignExtension"
   };
   v
}

qword SignExtension64 (w :: qword, size :: Zsize) =
   if size == Z64 then w else SignExtension (w, size, Z64)

nat maskShift (size :: Zsize, w :: qword) =
   if size == Z64 then [w<5:0>] else [w<4:0>]

qword ROL (size :: Zsize, x :: qword, y :: qword) =
   match size
   {
      case Z8 (_) => [x<7:0>  #<< ([y<4:0>]::nat)]
      case Z16    => [x<15:0> #<< ([y<4:0>]::nat)]
      case Z32    => [x<31:0> #<< ([y<4:0>]::nat)]
      case Z64    => x #<< ([y<5:0>]::nat)
   }

qword ROR (size :: Zsize, x :: qword, y :: qword) =
   match size
   {
      case Z8 (_) => [x<7:0>  #>> ([y<4:0>]::nat)]
      case Z16    => [x<15:0> #>> ([y<4:0>]::nat)]
      case Z32    => [x<31:0> #>> ([y<4:0>]::nat)]
      case Z64    => x #>> ([y<5:0>]::nat)
   }

qword SAR (size :: Zsize, x :: qword, y :: qword) =
   match size
   {
      case Z8 (_) => [x<7:0>  >> ([y<4:0>]::nat)]
      case Z16    => [x<15:0> >> ([y<4:0>]::nat)]
      case Z32    => [x<31:0> >> ([y<4:0>]::nat)]
      case Z64    => x >> ([y<5:0>]::nat)
   }

unit write_binop
   ( s :: Zsize,
     bop :: Zbinop_name,
     x :: qword,
     y :: qword,
     ea :: Zea) =
   match bop
   {
      case Zadd  => write_arith_result (s, add_with_carry_out (s, x, y), ea)
      case Zsub  => write_arith_result (s, sub_with_borrow (s, x, y), ea)
      case Zcmp  => write_arith_eflags (s, sub_with_borrow (s, x, y))
      case Ztest => write_logical_eflags (s, x && y)
      case Zand  => write_logical_result (s, x && y, ea)
      case Zxor  => write_logical_result (s, x ?? y, ea)
      case Zor   => write_logical_result (s, x || y, ea)
      case Zrol  => write_result_erase_eflags (ROL (s, x, y), ea)
      case Zror  => write_result_erase_eflags (ROR (s, x, y), ea)
      case Zsar  => write_result_erase_eflags (SAR (s, x, y), ea)
      case Zshl  => write_result_erase_eflags (x << maskShift (s, y), ea)
      case Zshr  => write_result_erase_eflags (x >>+ maskShift (s, y), ea)
      case Zadc  =>
         {
            carry = CF;
            result = x + y + [carry];
            CF <- value_width (s) <= [x] + [y] + [carry] :: nat;
            FlagUnspecified (Z_OF);
            write_arith_result_no_CF_OF (s, result, ea)
         }
      case Zsbb  =>
         {
            carry = CF;
            result = x - (y + [carry]);
            CF <- [x] < [y] + [carry] :: nat;
            FlagUnspecified (Z_OF);
            write_arith_result_no_CF_OF (s, result, ea)
         }
      -- rcl and rcr
      case _ => #FAILURE ("Binary op not implemented: " : [bop])
   }

-- Mon-ops

unit write_monop
   ( s :: Zsize,
     mop :: Zmonop_name,
     x :: qword,
     ea :: Zea) =
   match mop
   {
      case Znot => EA (ea) <- ~x
      case Zdec => write_arith_result_no_CF_OF (s, x - 1, ea)
      case Zinc => write_arith_result_no_CF_OF (s, x + 1, ea)
      case Zneg =>
         {
            write_arith_result_no_CF_OF (s, -x, ea);
            FlagUnspecified (Z_CF)
         }
   }

-- Bit tests

unit bit_test (bt :: Zbit_test_name, base::Zea, offset::nat) =
{
   bit = EA (base)<offset>;
   CF <- bit;
   match bt {
     case Zbt => nothing
     case Zbtc => EA (base)<offset> <- not bit
     case Zbtr => EA (base)<offset> <- false
     case Zbts => EA (base)<offset> <- true
   }
}

-- Evaluating conditions of eflags

bool read_cond (c :: Zcond) =
   match c
   {
      -- [A] above, [NBE] not below or equal (CF=0 and ZF=0)
      case Z_A  =>
         match EFLAGS (Z_CF), EFLAGS (Z_ZF)
         {
            case Some (false), Some (false) => true
            case Some (true), _ => false
            case _, Some (true) => false
            case _ => #BadFlagAccess("read_cond: " : [c])
         }

      -- [AE] above or equal, [NC] not carry, [NB] not below (CF=0)
      case Z_NB => not CF

      -- [B] below, [C] carry, [NAE] not above or equal (CF=1)
      case Z_B  => CF

      -- [BE] below or equal, [NA] not above (CF=1 or ZF=1)
      case Z_NA => -- CF or ZF
         match EFLAGS (Z_CF), EFLAGS (Z_ZF)
         {
            case Some (true), _ => true
            case _, Some (true) => true
            case Some (false), Some (false) => false
            case _ => #BadFlagAccess("read_cond: " : [c])
         }

      -- [E] equal, [Z] zero (ZF=1)
      case Z_E  => ZF

      -- [G] greater, [NLE] not less equal (ZF=0 and SF=OF)
      case Z_G  => -- not ZF and SF == OF
         match EFLAGS (Z_SF), EFLAGS (Z_OF)
         {
            case Some (a), Some (b) => a == b and not ZF
            case _ =>
               match EFLAGS (Z_ZF)
               {
                  case Some (true) => false
                  case _ => #BadFlagAccess("read_cond: " : [c])
               }
         }

      -- [GE] greater equal, [NL] not less (SF=OF)
      case Z_NL => SF == OF

      -- [L] less, [NGE] not greater equal (SF<>OF)
      case Z_L  => SF != OF

      -- [LE] less equal, [NG] not greater (ZF=1 or SF<>OF)
      case Z_NG => -- ZF or SF != OF
         match EFLAGS (Z_SF), EFLAGS (Z_OF)
         {
            case Some (a), Some (b) => a != b or ZF
            case _ =>
               match EFLAGS (Z_ZF)
               {
                  case Some (true) => true
                  case _ => #BadFlagAccess("read_cond: " : [c])
               }
         }

      -- [NE] not equal, [NZ] not zero (ZF=0)
      case Z_NE => not ZF

      -- [NO] not overflow (OF=0)
      case Z_NO => not OF

      -- [NP] not parity, [PO] parity odd (PF=0)
      case Z_NP => not PF

      -- [NS] not sign (SF=0)
      case Z_NS => not SF

      -- [O] overflow (OF=1)
      case Z_O  => OF

      -- [P] parity, [PE] parity even (PF=1)
      case Z_P  => PF

      -- [S] sign (SF=1)
      case Z_S  => SF

      case Z_ALWAYS => true
   }

-- Stack operations

qword x64_pop_aux =
{
   rsp = REG (RSP);
   top = mem64 (rsp);
   REG (RSP) <- rsp + 8;
   top
}

unit x64_pop (rm :: Zrm) = EA (ea_Zrm (Z64, rm)) <- x64_pop_aux

unit x64_pop_rip = RIP <- x64_pop_aux

unit x64_push_aux (w :: qword) =
{
   rsp = REG (RSP) - 8;
   REG (RSP) <- rsp;
   mem64 (rsp) <- w
}

unit x64_push (imm_rm :: Zimm_rm) = x64_push_aux (EA (ea_Zimm_rm (imm_rm)))

unit x64_push_rip = x64_push_aux (RIP)

unit x64_drop (imm :: qword) =
{
   when imm<7:0> != 0 do #FAILURE ("x64_drop");
   REG (RSP) <- REG (RSP) + imm
}

-- Floating-point

ieee_flags initial_ieee_flags (invalid :: bool) =
{
  var flags :: ieee_flags;
  flags.DivideByZero <- false;
  flags.InvalidOp <- invalid;
  flags.Overflow <- false;
  flags.Precision <- false;
  flags.Underflow <- false;
  flags
}

ieee_flags set_precision (flags :: ieee_flags, b :: bool) =
  { var f = flags; f.Precision <- b; f }

dword zero32 (a :: dword) = if Msb (a) then FP32_NegZero else FP32_PosZero
qword zero64 (a :: qword) = if Msb (a) then FP64_NegZero else FP64_PosZero

dword denormal_to_zero32 (a :: dword) =
  if MXCSR.DAZ and FP32_IsSubnormal (a) then zero32 (a) else a

qword denormal_to_zero64 (a :: qword) =
  if MXCSR.DAZ and FP64_IsSubnormal (a) then zero64 (a) else a

ieee_flags * dword flush_to_zero32 (flags :: ieee_flags, a :: dword) =
  if MXCSR.FZ and MXCSR.UM and flags.Underflow then
    (set_precision (flags, true), zero32 (a))
  else (flags, a)

ieee_flags * qword flush_to_zero64 (flags :: ieee_flags, a :: qword) =
  if MXCSR.FZ and MXCSR.UM and flags.Underflow then
    (set_precision (flags, true), zero64 (a))
  else (flags, a)

ieee_flags * qword sse_from_int64 (m :: ieee_rounding, w::bits(N)) =
{
  i = [w];
  q = FP64_FromInt (m, i);
  (set_precision (initial_ieee_flags (false), FP64_ToInt (m, q) != Some (i)), q)
}

ieee_flags * dword sse_from_int32 (m :: ieee_rounding, w::bits(N)) =
{
  i = [w];
  d = FP32_FromInt (m, i);
  (set_precision (initial_ieee_flags (false), FP32_ToInt (m, d) != Some (i)), d)
}

inline int INT_MIN (n::nat) = -(2 ** (n - 1))
inline int INT_MAX (n::nat) = 2 ** (n - 1) - 1

ieee_flags * bits(N) sse_to_int64 (m :: ieee_rounding, a::qword) =
  match FP64_ToInt (m, a)
  {
    case Some (i) =>
      if INT_MIN (N) <= i and i <= INT_MAX (N) then
      { w = [i];
        f = set_precision (initial_ieee_flags (false),
                           not (FP64_Equal (FP64_FromInt (m, [w]), a)));
        (f, w)
      }
      else
        (initial_ieee_flags (true), 1 #>> 1)
    case None => (initial_ieee_flags (true), 1 #>> 1)
  }

ieee_flags * bits(N) sse_to_int32 (m :: ieee_rounding, a::dword) =
  match FP32_ToInt (m, a)
  {
    case Some (i) =>
      if INT_MIN (N) <= i and i <= INT_MAX (N) then
      { w = [i];
        f = set_precision (initial_ieee_flags (false),
                           not (FP32_Equal (FP32_FromInt (m, [w]), a)));
        (f, w)
      }
      else
        (initial_ieee_flags (true), 1 #>> 1)
    case None => (initial_ieee_flags (true), 1 #>> 1)
  }

ieee_flags * dword float_min32 (a :: dword, b :: dword) =
{
  flags = initial_ieee_flags (FP32_Compare (a, b) == FP_UN);
  (flags, if not flags.InvalidOp and FP32_LessThan (a, b) then a else b)
}

ieee_flags * dword float_max32 (a :: dword, b :: dword) =
{
  flags = initial_ieee_flags (FP32_Compare (a, b) == FP_UN);
  (flags, if flags.InvalidOp or FP32_LessEqual (a, b) then b else a)
}

ieee_flags * qword float_min64 (a :: qword, b :: qword) =
{
  flags = initial_ieee_flags (FP64_Compare (a, b) == FP_UN);
  (flags, if not flags.InvalidOp and FP64_LessThan (a, b) then a else b)
}

ieee_flags * qword float_max64 (a :: qword, b :: qword) =
{
  flags = initial_ieee_flags (FP64_Compare (a, b) == FP_UN);
  (flags, if flags.InvalidOp or FP64_LessEqual (a, b) then b else a)
}

unit process_float_flags (l :: (bool * ieee_flags) list) =
{
  var denorm = false;
  var f = initial_ieee_flags (false);
  foreach x in l do
  {
    (d, flags) = x;
    denorm <- denorm or d;
    f.DivideByZero <- f.DivideByZero or flags.DivideByZero;
    f.InvalidOp <- f.InvalidOp or flags.InvalidOp;
    f.Overflow <- f.Overflow or flags.Overflow;
    f.Precision <- f.Precision or flags.Precision;
    f.Underflow <- f.Underflow or flags.Underflow
  };
  MXCSR.IE <- MXCSR.IE or f.InvalidOp;
  MXCSR.DE <- MXCSR.DE or denorm;
  MXCSR.ZE <- MXCSR.ZE or f.DivideByZero;
  when f.InvalidOp and not MXCSR.IM or
       denorm and not MXCSR.DM or
       f.DivideByZero and not MXCSR.ZM do
    XM_exception;
  MXCSR.OE <- MXCSR.OE or f.Overflow;
  MXCSR.PE <- MXCSR.PE or f.Precision;
  MXCSR.UE <- MXCSR.UE or f.Underflow;
  when f.Overflow and not MXCSR.OM or
       f.Precision and not MXCSR.PM or
       f.Underflow and not MXCSR.UM do
    XM_exception
}

ieee_rounding RoundingMode =
  match MXCSR.RC {
    case '00' => roundTiesToEven
    case '01' => roundTowardNegative
    case '10' => roundTowardPositive
    case '11' => roundTowardZero
  }

construct sse_binop { sse_add, sse_sub, sse_mul, sse_div, sse_max, sse_min }
construct sse_logic { sse_and, sse_andn, sse_or, sse_xor }
construct sse_compare {
  sse_eq_oq, sse_lt_os, sse_le_os, sse_unord_q,
  sse_neq_uq, sse_nlt_us, sse_nle_us, sse_ord_q
}

(bool * ieee_flags) * dword sse_binop32
  (bop :: sse_binop, a :: dword, b :: dword) =
{
  denorm = FP32_IsSubnormal (a) or FP32_IsSubnormal (b);
  a = denormal_to_zero32 (a);
  b = denormal_to_zero32 (b);
  (f, r) =
    match bop
    {
      case sse_add => flush_to_zero32 (FP32_Add_ (RoundingMode, a, b))
      case sse_sub => flush_to_zero32 (FP32_Sub_ (RoundingMode, a, b))
      case sse_mul => flush_to_zero32 (FP32_Mul_ (RoundingMode, a, b))
      case sse_div => flush_to_zero32 (FP32_Div_ (RoundingMode, a, b))
      case sse_max => float_max32 (a, b)
      case sse_min => float_min32 (a, b)
    };
  ((denorm, f), r)
}

(bool * ieee_flags) * qword sse_binop64
  (bop :: sse_binop, a :: qword, b :: qword) =
{
  denorm = FP64_IsSubnormal (a) or FP64_IsSubnormal (b);
  a = denormal_to_zero64 (a);
  b = denormal_to_zero64 (b);
  (f, r) =
    match bop
    {
      case sse_add => flush_to_zero64 (FP64_Add_ (RoundingMode, a, b))
      case sse_sub => flush_to_zero64 (FP64_Sub_ (RoundingMode, a, b))
      case sse_mul => flush_to_zero64 (FP64_Mul_ (RoundingMode, a, b))
      case sse_div => flush_to_zero64 (FP64_Div_ (RoundingMode, a, b))
      case sse_max => float_max64 (a, b)
      case sse_min => float_min64 (a, b)
    };
  ((denorm, f), r)
}

(bool * ieee_flags) * dword sse_sqrt32 (a :: dword) =
{
  -- sqrt never underflows so flush_to_zero32 not needed
  (f, r) = FP32_Sqrt_ (RoundingMode, denormal_to_zero32 (a));
  ((FP32_IsSubnormal (a), f), r)
}

(bool * ieee_flags) * qword sse_sqrt64 (a :: qword) =
{
  -- sqrt never underflows so flush_to_zero64 not needed
  (f, r) = FP64_Sqrt_ (RoundingMode, denormal_to_zero64 (a));
  ((FP64_IsSubnormal (a), f), r)
}

bits(N) sse_logic (bop :: sse_logic, a :: bits(N), b :: bits(N)) =
  match bop
  {
    case sse_and  => a && b
    case sse_or   => a || b
    case sse_xor  => a ?? b
    case sse_andn => ~a && b
  }

bool sse_compare_signalling (cmp :: sse_compare) =
   cmp in set {sse_lt_os, sse_le_os, sse_nlt_us, sse_nle_us}

(bool * ieee_flags) * dword sse_compare32
  (cmp :: sse_compare, a :: dword, b :: dword) =
{
  unordered = FP32_Compare (a, b) == FP_UN;
  c =
    match cmp
    {
      case sse_eq_oq   => FP32_Equal (a, b)
      case sse_lt_os   => FP32_LessThan (a, b)
      case sse_le_os   => FP32_LessEqual (a, b)
      case sse_unord_q => unordered
      case sse_neq_uq  => unordered or not (FP32_Equal (a, b))
      case sse_nlt_us  => unordered or not (FP32_LessThan (a, b))
      case sse_nle_us  => unordered or not (FP32_LessEqual (a, b))
      case sss_ord_q   => not unordered
    };
  flags = initial_ieee_flags
            (FP32_IsSignallingNan (a) or FP32_IsSignallingNan (b) or
             unordered and sse_compare_signalling (cmp));
  ((FP32_IsSubnormal (a) or FP32_IsSubnormal (b), flags), SignExtend ([c]`1))
}

(bool * ieee_flags) * qword sse_compare64
  (cmp :: sse_compare, a :: qword, b :: qword) =
{
  unordered = FP64_Compare (a, b) == FP_UN;
  c =
    match cmp
    {
      case sse_eq_oq   => FP64_Equal (a, b)
      case sse_lt_os   => FP64_LessThan (a, b)
      case sse_le_os   => FP64_LessEqual (a, b)
      case sse_unord_q => unordered
      case sse_neq_uq  => unordered or not (FP64_Equal (a, b))
      case sse_nlt_us  => unordered or not (FP64_LessThan (a, b))
      case sse_nle_us  => unordered or not (FP64_LessEqual (a, b))
      case sss_ord_q   => not unordered
    };
  flags = initial_ieee_flags
            (FP64_IsSignallingNan (a) or FP64_IsSignallingNan (b) or
             unordered and sse_compare_signalling (cmp));
  ((FP64_IsSubnormal (a) or FP64_IsSubnormal (b), flags), SignExtend ([c]`1))
}

---------------------------------------------------------------------------

-- XMM access

construct xmm_mem
{
   xmm_reg :: xreg  -- XMM register
   xmm_mem :: maddr -- mem [2^{scale} * index + base + displacement]
}

xmm_mem rm_to_xmm_mem (rm :: Zrm) =
  match rm
  {
    case Zr (r) => xmm_reg ([r])
    case Zm (m) => xmm_mem (m)
  }

component XMM (xm :: xmm_mem) :: dqword
{
   value =
     match xm {
       case xmm_reg(r) => XMM_REG(r)
       case xmm_mem(m) => mem128(mem_addr(m))
     }
   assign dqw =
     match xm {
       case xmm_reg(r) => XMM_REG(r) <- dqw
       case xmm_mem(m) => mem128(mem_addr(m)) <- dqw
     }
}

unit CheckAlignedXMM (xm :: xmm_mem, n :: nat) =
  match xm {
    case xmm_reg(r) => nothing
    case xmm_mem(m) =>
      { a = mem_addr(m); when a >>+ n << n <> a do GP_exception }
  }

inline unit CheckAlignedDoubleQuadword(xm::xmm_mem) = CheckAlignedXMM (xm, 4)

---------------------------------------------------------------------------

-- SSE Instructions (basic selection)

-- SSE: Pentium 3 (1999)
-- SSE2: Pentium 4 (2001)
-- SSE3: Prescott Pentium 4 (2004)
-- SSSE3: Core 2 (2007)
-- SSE4: Penryn Core 2 (2007-08)

-- SSE2
define SSE > bin_PD (bop :: sse_binop, dst :: xreg, src :: xmm_mem) =
{
  dst = xmm_reg (dst);
  'q11 q12' = XMM (dst);
  'q21 q22' = XMM (src);
  (f1, r1) = sse_binop64 (bop, q11, q21);
  (f2, r2) = sse_binop64 (bop, q12, q22);
  process_float_flags (list {f1, f2});
  XMM (dst) <- r1 : r2
}

-- SSE
define SSE > bin_PS (bop :: sse_binop, dst :: xreg, src :: xmm_mem) =
{
  dst = xmm_reg (dst);
  'd11 d12 d13 d14' = XMM (dst);
  'd21 d22 d23 d24' = XMM (src);
  (f1, r1) = sse_binop32 (bop, d11, d21);
  (f2, r2) = sse_binop32 (bop, d12, d22);
  (f3, r3) = sse_binop32 (bop, d13, d23);
  (f4, r4) = sse_binop32 (bop, d14, d24);
  process_float_flags (list {f1, f2, f3, f4});
  XMM (dst) <- r1 : r2 : r3 : r4
}

-- SSE2
define SSE > bin_SD (bop :: sse_binop, dst :: xreg, src :: xmm_mem) =
{
  dst = xmm_reg (dst);
  '_ q1' = XMM (dst);
  '_ q2' = XMM (src);
  (f, r) = sse_binop64 (bop, q1, q2);
  process_float_flags (list {f});
  XMM (dst)<63:0> <- r
}

-- SSE
define SSE > bin_SS (bop :: sse_binop, dst :: xreg, src :: xmm_mem) =
{
  dst = xmm_reg (dst);
  '_ d1' = XMM (dst);
  '_ d2' = XMM (src);
  (f, r) = sse_binop32 (bop, d1, d2);
  process_float_flags (list {f});
  XMM (dst)<31:0> <- r
}

-- SSE2
define SSE > logic_PD (bop :: sse_logic, dst :: xreg, src :: xmm_mem) =
{
  dst = xmm_reg (dst);
  'q11 q12' = XMM (dst);
  'q21 q22' = XMM (src);
  r1`64 = sse_logic (bop, q11, q21);
  r2`64 = sse_logic (bop, q12, q22);
  XMM (dst) <- r1 : r2
}

-- SSE
define SSE > logic_PS (bop :: sse_logic, dst :: xreg, src :: xmm_mem) =
{
  dst = xmm_reg (dst);
  'd11 d12 d13 d14' = XMM (dst);
  'd21 d22 d23 d24' = XMM (src);
  r1`32 = sse_logic (bop, d11, d21);
  r2`32 = sse_logic (bop, d12, d22);
  r3`32 = sse_logic (bop, d13, d23);
  r4`32 = sse_logic (bop, d14, d24);
  XMM (dst) <- r1 : r2 : r3 : r4
}

-- SSE2
define SSE > CMPPD (cmp :: sse_compare, dst :: xreg, src :: xmm_mem) =
{
  dst = xmm_reg (dst);
  'q11 q12' = XMM (dst);
  'q21 q22' = XMM (src);
  (f1, r1) = sse_compare64 (cmp, q11, q21);
  (f2, r2) = sse_compare64 (cmp, q12, q22);
  process_float_flags (list {f1, f2});
  XMM (dst) <- r1 : r2
}

-- SSE
define SSE > CMPPS (cmp :: sse_compare, dst :: xreg, src :: xmm_mem) =
{
  dst = xmm_reg (dst);
  'd11 d12 d13 d14' = XMM (dst);
  'd21 d22 d23 d24' = XMM (src);
  (f1, r1) = sse_compare32 (cmp, d11, d21);
  (f2, r2) = sse_compare32 (cmp, d12, d22);
  (f3, r3) = sse_compare32 (cmp, d13, d23);
  (f4, r4) = sse_compare32 (cmp, d14, d24);
  process_float_flags (list {f1, f2, f3, f4});
  XMM (dst) <- r1 : r2 : r3 : r4
}

-- SSE2
define SSE > CMPSD (cmp :: sse_compare, dst :: xreg, src :: xmm_mem) =
{
  dst = xmm_reg (dst);
  '_ q1' = XMM (dst);
  '_ q2' = XMM (src);
  (f, r) = sse_compare64 (cmp, q1, q2);
  process_float_flags (list {f});
  XMM (dst)<63:0> <- r
}

-- SSE
define SSE > CMPSS (cmp :: sse_compare, dst :: xreg, src :: xmm_mem) =
{
  dst = xmm_reg (dst);
  '_ q1' = XMM (dst);
  '_ q2' = XMM (src);
  (f, r) = sse_compare32 (cmp, q1, q2);
  process_float_flags (list {f});
  XMM (dst)<31:0> <- r
}

-- SSE2
define SSE > COMISD (src1 :: xreg, src2 :: xmm_mem) =
{
  '_ q1' = XMM (xmm_reg (src1));
  '_ q2' = XMM (src2);
  match FP64_Compare (q1, q2)
  {
    case FP_UN => { ZF <- true;  PF <- true;  CF <- true }
    case FP_GT => { ZF <- false; PF <- false; CF <- false }
    case FP_LT => { ZF <- false; PF <- false; CF <- true }
    case FP_EQ => { ZF <- true;  PF <- false; CF <- false }
  };
  OF <- false;
  AF <- false;
  SF <- false
}

-- SSE
define SSE > COMISS (src1 :: xreg, src2 :: xmm_mem) =
{
  '_ d1' = XMM (xmm_reg (src1));
  '_ d2' = XMM (src2);
  match FP32_Compare (d1, d2)
  {
    case FP_UN => { ZF <- true;  PF <- true;  CF <- true }
    case FP_GT => { ZF <- false; PF <- false; CF <- false }
    case FP_LT => { ZF <- false; PF <- false; CF <- true }
    case FP_EQ => { ZF <- true;  PF <- false; CF <- false }
  };
  OF <- false;
  AF <- false;
  SF <- false
}

-- SSE2
define SSE > CVTDQ2PD (dst :: xreg, src :: xmm_mem) =
{
  '_ i1`32 i2`32' = XMM (src);
  mode = RoundingMode;
  -- Precision exception not possible as all values are representable
  XMM (xmm_reg (dst)) <- FP64_FromInt (mode, [i1]) : FP64_FromInt (mode, [i2])
}

-- SSE2
define SSE > CVTDQ2PS (dst :: xreg, src :: xmm_mem) =
{
  'i1`32 i2`32 i3`32 i4' = XMM (src);
  mode = RoundingMode;
  f1, d1 = sse_from_int32 (mode, i1);
  f2, d2 = sse_from_int32 (mode, i2);
  f3, d3 = sse_from_int32 (mode, i3);
  f4, d4 = sse_from_int32 (mode, i4);
  process_float_flags
    (list {(false, f1), (false, f2), (false, f3), (false, f4)});
  XMM (xmm_reg (dst)) <- d1 : d2 : d3 : d4
}

-- SSE2
define SSE > CVTPD2DQ (truncate :: bool, dst :: xreg, src :: xmm_mem) =
{
  'q1 q2' = XMM (src);
  mode = if truncate then roundTowardZero else RoundingMode;
  f1, w1`32 = sse_to_int64 (mode, q1);
  f2, w2`32 = sse_to_int64 (mode, q2);
  process_float_flags (list {(false, f1), (false, f2)});
  XMM (xmm_reg (dst))<63:0> <- w1 : w2
}

-- SSE2
define SSE > CVTPD2PS (dst :: xreg, src :: xmm_mem) =
{
  'q1 q2' = XMM (src);
  mode = RoundingMode;
  f1, r1 = FP64_ToFP32_ (mode, q1);
  f2, r2 = FP64_ToFP32_ (mode, q1);
  f1 = FP64_IsSubnormal (q1), f1;
  f2 = FP64_IsSubnormal (q2), f2;
  process_float_flags (list {f1, f2});
  XMM (xmm_reg (dst))<63:0> <- r1 : r2
}

-- SSE2
define SSE > CVTPS2DQ (truncate :: bool, dst :: xreg, src :: xmm_mem) =
{
  'd1 d2 d3 d4' = XMM (src);
  mode = if truncate then roundTowardZero else RoundingMode;
  f1, w1`32 = sse_to_int32 (mode, d1);
  f2, w2`32 = sse_to_int32 (mode, d2);
  f3, w3`32 = sse_to_int32 (mode, d3);
  f4, w4`32 = sse_to_int32 (mode, d4);
  process_float_flags
    (list {(false, f1), (false, f2), (false, f3), (false, f4)});
  XMM (xmm_reg (dst)) <- w1 : w2 : w3 : w4
}

-- SSE2
define SSE > CVTPS2PD (dst :: xreg, src :: xmm_mem) =
{
  '_ d1 d2' = XMM (src);
  f1 = (FP32_IsSubnormal (d1), initial_ieee_flags (FP32_IsSignallingNan (d1)));
  f2 = (FP32_IsSubnormal (d2), initial_ieee_flags (FP32_IsSignallingNan (d2)));
  process_float_flags (list {f1, f2});
  XMM (xmm_reg (dst)) <- FP32_ToFP64 (d1) : FP32_ToFP64 (d2)
}

-- SSE2
define SSE > CVTSD2SI
  (truncate :: bool, quad :: bool, dst :: Zreg, src :: xmm_mem) =
{
  '_ q' = XMM (src);
  mode = if truncate then roundTowardZero else RoundingMode;
  if quad then
  {
    f, w`64 = sse_to_int64 (mode, q);
    process_float_flags (list {(false, f)});
    EA (Zea_r (Z64, dst)) <- w
  }
  else
  {
    f, w`32 = sse_to_int64 (mode, q);
    process_float_flags (list {(false, f)});
    EA (Zea_r (Z32, dst)) <- ZeroExtend (w)
  }
}

-- SSE2
define SSE > CVTSD2SS (dst :: xreg, src :: xmm_mem) =
{
  '_ q' = XMM (src);
  f, r = FP64_ToFP32_ (RoundingMode, q);
  f = FP64_IsSubnormal (q), f;
  process_float_flags (list {f});
  XMM (xmm_reg (dst))<31:0> <- r
}

-- SSE2
define SSE > CVTSI2SD (quad :: bool, dst :: xreg, src :: Zrm) =
{
  ea_src = ea_Zrm ((if quad then Z64 else Z32), src);
  f, q = sse_from_int64 (RoundingMode, EA (ea_src));
  process_float_flags (list {(false, f)});
  XMM (xmm_reg (dst))<63:0> <- q
}

-- SSE
define SSE > CVTSI2SS (quad :: bool, dst :: xreg, src :: Zrm) =
{
  ea_src = ea_Zrm ((if quad then Z64 else Z32), src);
  f, d = sse_from_int32 (RoundingMode, EA (ea_src));
  process_float_flags (list {(false, f)});
  XMM (xmm_reg (dst))<31:0> <- d
}

-- SSE2
define SSE > CVTSS2SD (dst :: xreg, src :: xmm_mem) =
{
  '_ d' = XMM (src);
  f = (FP32_IsSubnormal (d), initial_ieee_flags (FP32_IsSignallingNan (d)));
  process_float_flags (list {f});
  XMM (xmm_reg (dst))<63:0> <- FP32_ToFP64 (d)
}

-- SSE
define SSE > CVTSS2SI
  (truncate :: bool, quad :: bool, dst :: Zreg, src :: xmm_mem) =
{
  '_ d' = XMM (src);
  mode = if truncate then roundTowardZero else RoundingMode;
  if quad then
  {
    f, w`64 = sse_to_int32 (mode, d);
    process_float_flags (list {(false, f)});
    EA (Zea_r (Z64, dst)) <- w
  }
  else
  {
    f, w`32 = sse_to_int32 (mode, d);
    process_float_flags (list {(false, f)});
    EA (Zea_r (Z32, dst)) <- ZeroExtend (w)
  }
}

-- SSE2 and SSE
define SSE > MOVAP_D_S (double :: bool, dst :: xmm_mem, src :: xmm_mem) =
{
  CheckAlignedDoubleQuadword(dst);
  CheckAlignedDoubleQuadword(src);
  XMM (dst) <- XMM (src)
}

-- SSE2 and SSE
define SSE > MOVUP_D_S (double :: bool, dst :: xmm_mem, src :: xmm_mem) =
  XMM (dst) <- XMM (src)

-- SSE2
define SSE > MOV_D_Q (to_xmm :: bool, quad :: bool, xmm :: xreg, rm :: Zrm) =
  match to_xmm, quad
  {
     case false, false =>
       EA (ea_Zrm (Z32, rm)) <- ZeroExtend (XMM (xmm_reg (xmm))<31:0>)
     case false, true =>
       EA (ea_Zrm (Z64, rm)) <- XMM (xmm_reg (xmm))<63:0>
     case true, false =>
       XMM (xmm_reg (xmm)) <- ZeroExtend (EA (ea_Zrm (Z32, rm))<31:0>)
     case true, true =>
       XMM (xmm_reg (xmm)) <- ZeroExtend (EA (ea_Zrm (Z64, rm)))
  }

-- SSE2
define SSE > MOVQ (dst :: xmm_mem, src :: xmm_mem) =
  match dst
  {
     case xmm_reg (_) => XMM (dst) <- ZeroExtend (XMM (src)<63:0>)
     case xmm_mem (m) => EA (ea_Zrm (Z64, Zm (m))) <- XMM (src)<63:0>
  }

{-
-- MOVHPS is the same
define SSE > MOVHPD (dst :: xmm_mem, src :: xmm_mem) =
  XMM (dst)<127:64> <- XMM (src)<63:0>

-- MOVLPS is the same
define SSE > MOVLPD (dst :: xmm_mem, src :: xmm_mem) =
  XMM (dst)<63:0> <- XMM (src)<63:0>
-}

-- SSE2
define SSE > MOVSD (dst :: xmm_mem, src :: xmm_mem) =
  match src
  {
     case xmm_reg (_) => XMM (dst)<63:0> <- XMM (src)<63:0>
     case xmm_mem (_) => XMM (dst) <- ZeroExtend (XMM (src)<63:0>)
  }

-- SSE
define SSE > MOVSS (dst :: xmm_mem, src :: xmm_mem) =
  match src
  {
     case xmm_reg (_) => XMM (dst)<31:0> <- XMM (src)<31:0>
     case xmm_mem (_) => XMM (dst) <- ZeroExtend (XMM (src)<31:0>)
  }

-- SSE4.1
define SSE > PCMPEQQ (dst :: xreg, src :: xmm_mem) =
{
  dst = xmm_reg (dst);
  'q11 q12' = XMM (dst);
  'q21 q22' = XMM (src);
  XMM(dst) <- (if q11 == q21`64 then -1`64 else 0) :
              (if q12 == q22    then -1    else 0)
}

-- SSE2
define SSE > PSLLDQ (dst :: xreg, imm :: byte) =
{
  dst = xmm_reg (dst);
  XMM (dst) <- XMM (dst) << (Min (0n16, [imm]) * 8)
}

-- SSE2
define SSE > PSLLD_imm (dst :: xreg, imm :: byte) =
{
  dst = xmm_reg (dst);
  'd1`32 d2`32 d3`32 d4' = XMM (dst);
  i = [imm] :: nat;
  XMM (dst) <- (d1 << i) : (d2 << i) : (d3 << i) : (d4 << i)
}

-- SSE2
define SSE > PSLLQ_imm (dst :: xreg, imm :: byte) =
{
  dst = xmm_reg (dst);
  'q1`64 q2' = XMM (dst);
  i = [imm] :: nat;
  XMM (dst) <- (q1 << i) : (q2 << i)
}

-- SSE2
define SSE > PSLLW_imm (dst :: xreg, imm :: byte) =
{
  dst = xmm_reg (dst);
  'w1`16 w2`16 w3`16 w4`16 w5`16 w6`16 w7`16 w8' = XMM (dst);
  i = [imm] :: nat;
  XMM (dst) <- (w1 << i) : (w2 << i) : (w3 << i) : (w4 << i) :
               (w5 << i) : (w6 << i) : (w7 << i) : (w8 << i)
}

-- SSE2
define SSE > PSRAD_imm (dst :: xreg, imm :: byte) =
{
  dst = xmm_reg (dst);
  'd1`32 d2`32 d3`32 d4' = XMM (dst);
  i = [imm] :: nat;
  XMM (dst) <- (d1 >> i) : (d2 >> i) : (d3 >> i) : (d4 >> i)
}

-- SSE2
define SSE > PSRAW_imm (dst :: xreg, imm :: byte) =
{
  dst = xmm_reg (dst);
  'w1`16 w2`16 w3`16 w4`16 w5`16 w6`16 w7`16 w8' = XMM (dst);
  i = [imm] :: nat;
  XMM (dst) <- (w1 >> i) : (w2 >> i) : (w3 >> i) : (w4 >> i) :
               (w5 >> i) : (w6 >> i) : (w7 >> i) : (w8 >> i)
}

-- SSE2
define SSE > PSRLDQ (dst :: xreg, imm :: byte) =
{
  dst = xmm_reg (dst);
  XMM (dst) <- XMM (dst) >>+ (Min (0n16, [imm]) * 8)
}

-- SSE2
define SSE > PSRLD_imm (dst :: xreg, imm :: byte) =
{
  dst = xmm_reg (dst);
  'd1`32 d2`32 d3`32 d4' = XMM (dst);
  i = [imm] :: nat;
  XMM (dst) <- (d1 >>+ i) : (d2 >>+ i) : (d3 >>+ i) : (d4 >>+ i)
}

-- SSE2
define SSE > PSRLQ_imm (dst :: xreg, imm :: byte) =
{
  dst = xmm_reg (dst);
  'q1`64 q2' = XMM (dst);
  i = [imm] :: nat;
  XMM (dst) <- (q1 >>+ i) : (q2 >>+ i)
}

-- SSE2
define SSE > PSRLW_imm (dst :: xreg, imm :: byte) =
{
  dst = xmm_reg (dst);
  'w1`16 w2`16 w3`16 w4`16 w5`16 w6`16 w7`16 w8' = XMM (dst);
  i = [imm] :: nat;
  XMM (dst) <- (w1 >>+ i) : (w2 >>+ i) : (w3 >>+ i) : (w4 >>+ i) :
               (w5 >>+ i) : (w6 >>+ i) : (w7 >>+ i) : (w8 >>+ i)
}

-- SSE2
define SSE > SQRTPD (dst :: xreg, src :: xmm_mem) =
{
  'q1 q2' = XMM (src);
  (f1, r1) = sse_sqrt64 (q1);
  (f2, r2) = sse_sqrt64 (q2);
  process_float_flags (list {f1, f2});
  XMM (xmm_reg (dst)) <- r1 : r2
}

-- SSE2
define SSE > SQRTSD (dst :: xreg, src :: xmm_mem) =
{
  '_ q' = XMM (src);
  (f, r) = sse_sqrt64 (q);
  process_float_flags (list {f});
  XMM (xmm_reg (dst))<63:0> <- r
}

-- SSE
define SSE > SQRTPS (dst :: xreg, src :: xmm_mem) =
{
  'd1 d2 d3 d4' = XMM (src);
  (f1, r1) = sse_sqrt32 (d1);
  (f2, r2) = sse_sqrt32 (d2);
  (f3, r3) = sse_sqrt32 (d3);
  (f4, r4) = sse_sqrt32 (d4);
  process_float_flags (list {f1, f2, f3, f4});
  XMM (xmm_reg (dst)) <- r1 : r2 : r3 : r4
}

-- SSE
define SSE > SQRTSS (dst :: xreg, src :: xmm_mem) =
{
  '_ d' = XMM (src);
  (f, r) = sse_sqrt32 (d);
  process_float_flags (list {f});
  XMM (xmm_reg (dst))<31:0> <- r
}

---------------------------------------------------------------------------

-- =====================
-- Operational Semantics
-- =====================

define Zbinop (bop :: Zbinop_name, size :: Zsize, dst_src :: Zdest_src) =
{
   ea, val_dst, val_src = read_dest_src_ea (size, dst_src);
   write_binop (size, bop, val_dst, val_src, ea)
}

define Zbit_test (bt :: Zbit_test_name, size :: Zsize, dst_src :: Zdest_src) =
{
   ea_src = ea_Zsrc (size, dst_src);
   offset = EA (ea_src);
   match dst_src {
     case Zr_rm (_) => UD_exception
     case Zrm_r (Zr (_), _) or Zrm_i (Zr (_), _) =>
       bit_test (bt, ea_Zdest (size, dst_src), [modSize (size, offset)])
     case Zrm_i (Zm (si, base, disp), _) =>
     {
       offset = modSize (size, offset);
       bit_base = ea_Zrm (Z8 (false), Zm (si, base, disp + (offset sdiv 8)));
       bit_test (bt, bit_base, [offset smod 8])
     }
     case Zrm_r (Zm (si, base, disp), _) =>
     {
       offset = SignExtension64 (offset, size);
       bit_base = ea_Zrm (Z8 (false), Zm (si, base, disp + (offset sdiv 8)));
       bit_test (bt, bit_base, [offset smod 8])
     }
   }
}

define Zcall (imm_rm :: Zimm_rm) =
{
   x64_push_rip;
   jump_to_ea (ea_Zimm_rm (imm_rm))
}

define Zcmpxchg (size :: Zsize, rm :: Zrm, r :: Zreg) =
{
   ea_src = Zea_r (size, r);
   ea_acc = Zea_r (size, RAX);
   ea_dst = ea_Zrm (size, rm);
   val_dst = EA (ea_dst);
   acc = EA (ea_src);
   write_binop (size, Zcmp, acc, val_dst, ea_src);
   if acc == val_dst
      then EA (ea_dst) <- EA (ea_src)
   else EA (ea_acc) <- val_dst
}

define Zdiv (size :: Zsize, rm :: Zrm) =
{
   w = value_width (size);
   ea_eax = Zea_r (size, RAX);
   ea_edx = Zea_r (size, RDX);
   ea_rm = ea_Zrm (size, rm);
   n = [EA (ea_edx)] * w + [EA (ea_eax)] :: nat;
   d = [EA (ea_rm)] :: nat;
   q = n div d;
   r = n mod d;
   when d == 0 or w <= q do DE_exception;
   EA (ea_eax) <- [q];
   EA (ea_edx) <- [r];
   erase_eflags
}

define Zidiv (size :: Zsize, rm :: Zrm) =
{
   w = [value_width (size)] :: int;
   ea_eax = Zea_r (size, RAX);
   ea_edx = Zea_r (size, RDX);
   ea_rm = ea_Zrm (size, rm);
   n = [SignExtension64 (EA (ea_edx), size)] * w + [[EA (ea_eax)]::nat];
   d = [SignExtension64 (EA (ea_rm),  size)] :: int;
   q = n quot d;
   r = n rem d;
   when d == 0 or q < -(w div 2) or w div 2 <= q do DE_exception;
   EA (ea_eax) <- [q];
   EA (ea_edx) <- [r];
   erase_eflags
}

-- includes jmp rel, i.e. unconditional relative jumps.
define Zjcc (cond :: Zcond, imm :: qword) =
   when read_cond (cond) do RIP <- RIP + imm

-- jmp excludes relative jumps, see jcc.
define Zjmp (rm :: Zrm) = RIP <- EA (ea_Zrm (Z64, rm))

define Zlea (size :: Zsize, dst_src :: Zdest_src) =
{
   ea_src = ea_Zsrc (size, dst_src);
   ea_dst = ea_Zdest (size, dst_src);
   EA (ea_dst) <- get_ea_address (ea_src)
}

define Zleave =
{
   REG(RSP) <- REG(RBP);
   x64_pop (Zr (RBP))
}

define Zloop (cond :: Zcond, imm :: qword) =
{
   ecx1 = REG (RCX) - 1;
   REG (RCX) <- ecx1;
   when ecx1 != 0 and read_cond (cond) do RIP <- RIP + imm
}

define Zmonop (mop :: Zmonop_name, size :: Zsize, rm :: Zrm) =
{
   ea = ea_Zrm (size, rm);
   write_monop (size, mop, EA (ea), ea)
}

define Zmov (cond :: Zcond, size :: Zsize, dst_src :: Zdest_src) =
   when read_cond (cond) do
   {
      ea_src = ea_Zsrc (size, dst_src);
      ea_dst = ea_Zdest (size, dst_src);
      EA (ea_dst) <- EA (ea_src)
   }

define Zmovsx (size1 :: Zsize, dst_src :: Zdest_src, size2 :: Zsize) =
   EA (ea_Zdest (size2, dst_src)) <-
      SignExtension (EA (ea_Zsrc (size1, dst_src)), size1, size2)

define Zmovzx (size1 :: Zsize, dst_src :: Zdest_src, size2 :: Zsize) =
   EA (ea_Zdest (size2, dst_src)) <- EA (ea_Zsrc (size1, dst_src))

define Zmul (size :: Zsize, rm :: Zrm) =
{
   ea_eax = Zea_r (size, RAX);
   eax = EA (ea_eax);
   val_src = EA (ea_Zrm (size, rm));
   match size
   {
      case Z8 (_) => EA (Zea_r (Z16, RAX)) <- eax * val_src
      case _ =>
      {
         EA (ea_eax) <- eax * val_src;
         ea_edx = Zea_r (size, RDX);
         EA (ea_edx) <- [([eax] * [val_src] :: nat) div value_width (size)]
      }
   };
   erase_eflags  -- over appoximation
}

define Zimul (size :: Zsize, rm :: Zrm) =
{
   ea_eax = Zea_r (size, RAX);
   eax = SignExtension64 (EA (ea_eax), size);
   val_src = SignExtension64 (EA (ea_Zrm (size, rm)), size);
   product = [eax] * [val_src] :: int;
   product64 = [product]`64;
   erase_eflags;
   overflow = not (if size == Z64 then [product64] == product
                   else SignExtension64 (product64, size) == product64);
   CF <- overflow;
   OF <- overflow;
   match size
   {
      case Z8 (_) => EA (Zea_r (Z16, RAX)) <- product64
      case _ =>
      {
         EA (ea_eax) <- product64;
         ea_edx = Zea_r (size, RDX);
         EA (ea_edx) <- [product div [value_width (size)]]
      }
   }
}

define Zimul2 (size :: Zsize, r :: Zreg, rm :: Zrm) =
{
   ea_dst = Zea_r (size, r);
   val_dst = SignExtension64 (EA (ea_dst), size);
   val_src = SignExtension64 (EA (ea_Zrm (size, rm)), size);
   product = [val_dst] * [val_src] :: int;
   product64 = [product]`64;
   erase_eflags;
   overflow = not (if size == Z64 then [product64] == product
                   else SignExtension64 (product64, size) == product64);
   CF <- overflow;
   OF <- overflow;
   EA (ea_dst) <- product64
}

define Zimul3 (size :: Zsize, r :: Zreg, rm :: Zrm, imm :: qword) =
{
   val_src = SignExtension64 (EA (ea_Zrm (size, rm)), size);
   product = [val_src] * [imm] :: int;
   product64 = [product]`64;
   erase_eflags;
   overflow = not (if size == Z64 then [product64] == product
                   else SignExtension64 (product64, size) == product64);
   CF <- overflow;
   OF <- overflow;
   EA (Zea_r (size, r)) <- product64
}

define Znop (n::nat) = ()

define Zpop (rm :: Zrm) = x64_pop (rm)

define Zpush (imm_rm :: Zimm_rm) = x64_push (imm_rm)

define Zret (imm :: qword) =
{
   x64_pop_rip;
   x64_drop (imm)
}

define Zset (cond :: Zcond, have_rex :: bool, rm :: Zrm) =
  EA (ea_Zrm (Z8 (have_rex), rm)) <- [read_cond (cond)]

define Zxadd (size :: Zsize, rm :: Zrm, r :: Zreg) =
{
   ea_src = Zea_r (size, r);
   ea_dst = ea_Zrm (size, rm);
   val_src = EA (ea_src);
   val_dst = EA (ea_dst);
   EA (ea_src) <- val_dst;
   write_binop (size, Zadd, val_src, val_dst, ea_dst)
}

define Zxchg (size :: Zsize, rm :: Zrm, r :: Zreg) =
{
   ea_src = Zea_r (size, r);
   ea_dst = ea_Zrm (size, rm);
   val_src = EA (ea_src);
   val_dst = EA (ea_dst);
   EA (ea_src) <- val_dst;
   EA (ea_dst) <- val_src
}

define Zcmc = CF <- not CF

define Zclc = CF <- false

define Zstc = CF <- true

define Run

---------------------------------------------------------------------------

-- ====================
-- Instruction Decoding
-- ====================

construct Zinst
{
   Zfull_inst :: stream * instruction * ostream
   Zdec_fail  :: string
}

-- Parse immediates

qword * ostream oimmediate8 (strm :: ostream) =
   match strm
   {
      case Some (b @ t) => SignExtend (b), Some (t)
      case _ => UNKNOWN, None
   }

qword * ostream immediate8 (strm :: stream) = oimmediate8 (Some (strm))

qword * ostream immediate16 (strm :: stream) =
   match strm
   {
      case b1 @ b2 @ t => SignExtend (b2 : b1 ), Some (t)
      case _ => UNKNOWN, None
   }

qword * ostream immediate32 (strm :: stream) =
   match strm
   {
      case b1 @ b2 @ b3 @ b4 @ t => SignExtend (b4 : b3 : b2 : b1), Some (t)
      case _ => UNKNOWN, None
   }

qword * ostream immediate64 (strm :: stream) =
   match strm
   {
      case b1 @ b2 @ b3 @ b4 @ b5 @ b6 @ b7 @ b8 @ t =>
        (b8 : b7 : b6 : b5 : b4 : b3 : b2 : b1), Some (t)
      case _ => UNKNOWN, None
   }

qword * ostream immediate (size :: Zsize, strm :: stream) =
   match size
   {
      case Z8 (_) => immediate8 (strm)
      case Z16    => immediate16 (strm)
      case _      => immediate32 (strm)
   }

qword * ostream oimmediate (size :: Zsize, strm :: ostream) =
  match strm
  {
    case Some (s) => immediate (size, s)
    case None => UNKNOWN, None
  }

qword * ostream full_immediate (size :: Zsize, strm :: stream) =
  if size == Z64 then immediate64 (strm) else immediate (size, strm)

---------------------------------------------------------------------------

-- Parse the ModRM and SIB bytes

register REX :: bits(4)
{
   3:W, -- 0 = Operand size dtermined by CS.D; 1 = 64-bit Operand Size
   2:R, -- Extension of the ModR/M reg field
   1:X, -- Extension of the ModR/M SIB (Scale,Index,Base) index field
   0:B  -- Extension of the ModR/M r/m field, SIB base field or opcode reg field
}

Zreg RexReg (b :: bool, r :: bits(3)) = [[b]`1 : r]

qword * ostream readDisplacement (Mod :: bits(2), strm :: stream) =
   if Mod == 1
      then immediate8 (strm)
   else if Mod == 2
      then immediate32 (strm)
   else (0, Some (strm))

qword * ostream readSibDisplacement (Mod :: bits(2), strm :: stream) =
   if Mod == 1 then immediate8 (strm) else immediate32 (strm)

Zrm * ostream readSIB (REX :: REX, Mod :: bits(2), strm :: stream) =
   match strm
   {  case 'SS Index Base' @ strm1 =>
         { base = RexReg (REX.B, Base);
           index = RexReg (REX.X, Index);
           scaled_index = if index == RSP then None else Some (SS, index);
           if base == RBP
              then {
                      displacement, strm2 = readSibDisplacement (Mod, strm1);
                      base = if Mod == 0 then ZnoBase else ZregBase (base);
                      (Zm (scaled_index, base, displacement), strm2)
                   }
           else
           {
              displacement, strm2 = readDisplacement (Mod, strm1);
              (Zm (scaled_index, ZregBase (base), displacement), strm2)
           }
         }
      case _ => (UNKNOWN, None)
   }

Zreg * Zrm * ostream readModRM (REX :: REX, strm :: stream) =
   match strm {
     case '00 RegOpc 101' @ strm1 =>
       { displacement, strm2 = immediate32 (strm1);
         (RexReg (REX.R, RegOpc), Zm (None, ZripBase, displacement), strm2)
       }
     case '11 REG RM' @ strm1 =>
         (RexReg (REX.R, REG), Zr (RexReg (REX.B, RM)), Some (strm1))
     case 'Mod RegOpc 100' @ strm1 =>
       { sib, strm2 = readSIB (REX, Mod, strm1);
         (RexReg (REX.R, RegOpc), sib, strm2)
       }
     case 'Mod RegOpc RM' @ strm1 =>
       { displacement, strm2 = readDisplacement (Mod, strm1);
         (RexReg (REX.R, RegOpc),
          Zm (None, ZregBase (RexReg (REX.B, RM)), displacement),
          strm2)
       }
     case _ => (UNKNOWN, UNKNOWN, None)
   }

bits(3) * Zrm * ostream readOpcodeModRM (REX :: REX, strm :: stream) =
{  opcode, rm, strm1 = readModRM (REX, strm);
   ([[opcode] mod 0n8], rm, strm1)
}

---------------------------------------------------------------------------

-- Parse Prefixes

nat prefixGroup (b :: byte) =
   match b
   { -- Lock, REPNE/REPNZ, REPE/REPZ prefixes
     -- F2 and F3 can be mandatory
     case 0xF0 or 0xF2 or 0xF3 => 1
     -- Segment override prefixes
     -- Branch Hints
     case 0x26 or 0x2E or 0x36 or 0x3E or 0x64 or 0x65 => 2
     -- Operand-size override
     -- can be mandatory
     case 0x66 => 3
     -- Address-size override
     case 0x67 => 4
     -- REX prefix
     case _ => if b<7:4> == '0100' then 5 else 0
   }

(stream * bool * REX * stream) option
   readPrefix (s :: nat set, p :: stream, strm :: stream) =
   match strm
   {  case h @ strm1 =>
         {  group = prefixGroup (h);
            if group == 0
               then Some (p, false, REX ('0000'), strm)
            else if group == 5
               then Some (p, true, REX (h<3:0>), strm1)
            else if group in s
               then None
            else readPrefix (group insert s, h @ p, strm1)
         }
     case Nil => Some (p, false, UNKNOWN, strm)
   }

(stream * bool * REX * stream) option readPrefixes (strm :: stream) =
   readPrefix (set {}, Nil, strm)

---------------------------------------------------------------------------

-- Operand Size

-- w from REX.W
-- v from opcode
-- override from 0x66 prefix

-- r/m8, imm8        not v
-- r/m16, imm16      override and not w and v
-- r/m32, imm32      not W and v
-- r/m64, imm32      w and v

Zsize OpSize (have_rex :: bool, w :: bool, v :: bits(1), override :: bool) =
  if v == 0
    then Z8 (have_rex)
  else if w
    then Z64
  else if override
    then Z16
  else Z32

bool isZm (rm :: Zrm) = match rm { case Zm (_) => true case _ => false }

---------------------------------------------------------------------------

-- The decoder

Zinst x64_decode (strm :: stream) =
match readPrefixes (strm)
{
  case None => Zdec_fail ("Bad prefix")
  case Some (p, have_rex, REX, strm1) =>
  {
    prefixes = SetOfList(p);
    op_size_override = 0x66 in prefixes;
    if 0x67 in prefixes
       then Zdec_fail ("address override prefix not supported")
    else
       match strm1
       { -- Binop (ADD..CMP)
         -- ADD r/mX, rX
         -- ADD rX, r/mX
         case '00 opc 0 x v' @ strm2 =>
            { reg, rm, strm3 = readModRM (REX, strm2);
              size = OpSize (have_rex, REX.W, v, op_size_override);
              binop = [opc] :: Zbinop_name;
              src_dst = if x == 0`1 then Zrm_r (rm, reg) else Zr_rm (reg, rm);
              Zfull_inst (p, Zbinop (binop, size, src_dst), strm3)
            }

         -- Binop (ADD..CMP)
         -- ADD EAX, immX
         case '00 opc 1 0 v' @ strm2 =>
            { size = OpSize (have_rex, REX.W, v, op_size_override);
              imm, strm3 = immediate (size, strm2);
              Zfull_inst (p, Zbinop ([opc], size, Zrm_i (Zr (RAX), imm)), strm3)
            }

         -- PUSH r/mX
         -- POP r/mX
         case '0x5 b r' @ strm2 =>
            { reg = Zr ([[REX.B] : r]);
              Zfull_inst
                (p, (if b == 0`1 then Zpush (Zrm (reg)) else Zpop (reg)),
                 Some (strm2))
            }

         -- MOVSXD r64, r/m32
         case 0x63 @ strm2 =>
            {  reg, rm, strm3 = readModRM (REX, strm2);
               Zfull_inst (p, Zmovsx (Z32, Zr_rm (reg, rm), Z64), strm3)
            }

         -- PUSH immX
         case '0x6 10 b 0' @ strm2 =>
            { imm, strm3 =
                if b == 1 then immediate8 (strm2) else immediate32 (strm2);
              Zfull_inst (p, Zpush (Zimm (imm)), strm3)
            }

         -- IMUL rx, r/mX, immX
         case '0x6 10 b 1' @ strm2 =>
            { size = OpSize (have_rex, REX.W, 1, op_size_override);
              reg, rm, strm3 = readModRM (REX, strm2);
              match strm3
              {
                case None => Zdec_fail ("not enough bytes")
                case Some (s) =>
                {
                  imm, strm4 =
                    if b == 1 then immediate8 (s) else immediate32 (s);
                  Zfull_inst (p, Zimul3 (size, reg, rm, imm), strm4)
                }
              }
            }

         -- Jcc rel8
         case '0x7 c' @ strm2 =>
            { imm, strm3 = immediate8 (strm2);
              Zfull_inst (p, Zjcc ([c], imm), strm3)
            }

         -- Immediate Group 1 (ADD..CMP)
         -- ADD r/mX, immX
         case '0x8 000 v' @ strm2 =>
            { size = OpSize (have_rex, REX.W, v, op_size_override);
              opcode, rm, strm3 = readOpcodeModRM (REX, strm2);
              imm, strm4 = oimmediate (size, strm3);
              binop = [opcode] :: Zbinop_name;
              Zfull_inst (p, Zbinop (binop, size, Zrm_i (rm, imm)), strm4)
            }
         -- ADD r/mX, imm8
         case 0x83 @ strm2 =>
            { size = OpSize (false, REX.W, 1, op_size_override);
              opcode, rm, strm3 = readOpcodeModRM (REX, strm2);
              imm, strm4 = oimmediate8 (strm3);
              binop = [opcode] :: Zbinop_name;
              Zfull_inst (p, Zbinop (binop, size, Zrm_i (rm, imm)), strm4)
            }

         -- TEST r/mX, rX
         case '0x8 010 v' @ strm2 =>
            { size = OpSize (have_rex, REX.W, v, op_size_override);
              reg, rm, strm3 = readModRM (REX, strm2);
              Zfull_inst (p, Zbinop (Ztest, size, Zrm_r (rm, reg)), strm3)
            }

         -- XCHG r/mX, rX
         case '0x8 011 v' @ strm2 =>
            { size = OpSize (have_rex, REX.W, v, op_size_override);
              reg, rm, strm3 = readModRM (REX, strm2);
              Zfull_inst (p, Zxchg (size, rm, reg), strm3)
            }

         -- MOV r/mX, rX
         -- MOV rX, r/mX
         case '0x8 10 x v' @ strm2 =>
            { reg, rm, strm3 = readModRM (REX, strm2);
              size = OpSize (have_rex, REX.W, v, op_size_override);
              src_dst = if x == 0`1 then Zrm_r (rm, reg) else Zr_rm (reg, rm);
              Zfull_inst (p, Zmov (Z_ALWAYS, size, src_dst), strm3)
            }

         -- LEA r/mX, m
         case '0x8D' @ strm2 =>
            { size = OpSize (true, REX.W, 1, op_size_override);
              reg, rm, strm3 = readModRM (REX, strm2);
              if isZm (rm)
                 then Zfull_inst (p, Zlea (size, Zr_rm (reg, rm)), strm3)
              else Zdec_fail ("LEA with register argument")
            }

         -- Unary Group 1a
         case '0x8F' @ strm2 =>
            { opcode, rm, strm3 = readOpcodeModRM (REX, strm2);
              if opcode == 0
                 then Zfull_inst (p, Zpop (rm), strm3)
              else Zdec_fail ("Unsupported opcode: Group 1a")
            }

         -- XCHG EAX, rX
         case '0x9 0 r' @ strm2 =>
            { size = OpSize (true, REX.W, 1, op_size_override);
              reg = RexReg (REX.B, r);
              if reg == RAX
                 then Zfull_inst (p, Znop (Length (strm)), Some (strm2))
              else Zfull_inst (p, Zxchg (size, Zr (RAX), reg), Some (strm2))
            }

         -- TEST EAX, immX
         case '0xA 100 v' @ strm2 =>
            { size = OpSize (true, REX.W, v, op_size_override);
              imm, strm3 = immediate (size, strm2);
              Zfull_inst (p, Zbinop (Ztest, size, Zrm_i (Zr (RAX), imm)), strm3)
            }

         -- MOV rX, immX
         case '0xB v r' @ strm2 =>
            { size = OpSize (have_rex, REX.W, v, op_size_override);
              imm, strm3 = full_immediate (size, strm2);
              reg = RexReg (REX.B, r);
              Zfull_inst
                 (p, Zmov (Z_ALWAYS, size, Zrm_i (Zr (reg), imm)), strm3)
            }

         -- Shift Group 2
         case '0xC 000 v' @ strm2 =>
            { size = OpSize (have_rex, REX.W, v, op_size_override);
              opcode, rm, strm3 = readOpcodeModRM (REX, strm2);
              imm, strm4 = oimmediate8 (strm3);
              binop = [[opcode] + 0n8] :: Zbinop_name;
              if opcode == '110'
                 then Zdec_fail ("Unsupported opcode: Shift Group 2")
              else Zfull_inst (p, Zbinop (binop, size, Zrm_i (rm, imm)), strm4)
            }

         -- RETN
         -- RETN imm16
         case '0xC 001 v' @ strm2 =>
            if v == 0
               then { imm, strm3 = immediate16 (strm2);
                      Zfull_inst (p, Zret (imm), strm3)
                    }
            else Zfull_inst (p, Zret (0), Some (strm2))

         -- Group 11
         case '0xC 011 v' @ strm2 =>
            { size = OpSize (have_rex, REX.W, v, op_size_override);
              opcode, rm, strm3 = readOpcodeModRM (REX, strm2);
              imm, strm4 = oimmediate (size, strm3);
              if opcode == '000'
                 then -- MOV r/mX, immX
                      Zfull_inst
                         (p, Zmov (Z_ALWAYS, size, Zrm_i (rm, imm)), strm4)
              else Zdec_fail ("Unsupported opcode: Group 11")
            }

         -- LEAVE
         case 0xC9 @ strm2 =>
            Zfull_inst (p, Zleave, Some (strm2))

         -- Shift Group 2
         case '0xD 00 b v' @ strm2 =>
            { size = OpSize (have_rex, REX.W, v, op_size_override);
              opcode, rm, strm3 = readOpcodeModRM (REX, strm2);
              shift = if b == 0`1 then Zrm_i (rm, 1) else Zrm_r (rm, RCX);
              binop = [[opcode] + 0n8] :: Zbinop_name;
              if opcode == '110'
                 then Zdec_fail ("Unsupported opcode: Shift Group 2")
              else Zfull_inst (p, Zbinop (binop, size, shift), strm3)
            }

         -- LOOPE rel8
         -- LOOPNE rel8
         case '0xE 000 b' @ strm2 =>
            { imm, strm3 = immediate8 (strm2);
              cond = if b == 0 then Z_NE else Z_E;
              Zfull_inst (p, Zloop (cond, imm), strm3)
            }

         -- LOOP rel8
         case '0xE2' @ strm2 =>
            { imm, strm3 = immediate8 (strm2);
              Zfull_inst (p, Zloop (Z_ALWAYS, imm), strm3)
            }

         -- CALL rel32
         case '0xE8' @ strm2 =>
            { imm, strm3 = immediate32 (strm2);
              Zfull_inst (p, Zcall (Zimm (imm)), strm3)
            }

         -- JMP rel8
         -- JMP rel32
         case '0xE 10 b 1' @ strm2 =>
            { imm, strm3 =
                if b == 0 then immediate32 (strm2) else immediate8 (strm2);
              Zfull_inst (p, Zjcc (Z_ALWAYS, imm), strm3)
            }

         case 0xF5 @ strm2 => Zfull_inst (p, Zcmc, Some (strm2))

         -- Unary Group 3
         case '0xF 011 v' @ strm2 =>
            { size = OpSize (have_rex, REX.W, v, op_size_override);
              opcode, rm, strm3 = readOpcodeModRM (REX, strm2);
              match opcode
              {  -- TEST r/mX, immX
                 case '000' =>
                  { imm, strm4 = oimmediate (size, strm3);
                    Zfull_inst (p, Zbinop (Ztest, size, Zrm_i (rm, imm)), strm4)
                  }

                 -- NOT r/mX
                 case '010' =>
                    Zfull_inst (p, Zmonop (Znot, size, rm), strm3)

                 -- NEG r/mX
                 case '011' =>
                    Zfull_inst (p, Zmonop (Zneg, size, rm), strm3)

                 -- MUL r/mX
                 case '100' =>
                    Zfull_inst (p, Zmul (size, rm), strm3)

                 -- IMUL r/mX
                 case '101' =>
                    Zfull_inst (p, Zimul (size, rm), strm3)

                 -- DIV r/mX
                 case '110' =>
                    Zfull_inst (p, Zdiv (size, rm), strm3)

                 -- IDIV r/mX
                 case '111' =>
                    Zfull_inst (p, Zidiv (size, rm), strm3)

                 case _ => Zdec_fail ("Unsupported opcode: Unary Group 3")
              }
            }

         case 0xF8 @ strm2 => Zfull_inst (p, Zclc, Some (strm2))
         case 0xF9 @ strm2 => Zfull_inst (p, Zstc, Some (strm2))

         -- INC/DEC Group 4
         case 0xFE @ strm2 =>
            { opcode, rm, strm3 = readOpcodeModRM (REX, strm2);
              if opcode == '000'
                 then Zfull_inst (p, Zmonop (Zinc, Z8 (have_rex), rm), strm3)
              else if opcode == '001'
                 then Zfull_inst (p, Zmonop (Zdec, Z8 (have_rex), rm), strm3)
              else Zdec_fail ("Unsupported opcode: INC/DEC Group 4")
            }

         -- INC/DEC Group 5
         case 0xFF @ strm2 =>
            { size = OpSize (have_rex, REX.W, 1, op_size_override);
              opcode, rm, strm3 = readOpcodeModRM (REX, strm2);
              match opcode
              {  -- INC r/mX
                 case '000' =>
                    Zfull_inst (p, Zmonop (Zinc, size, rm), strm3)
                 -- DEC rm/X
                 case '001' =>
                    Zfull_inst (p, Zmonop (Zdec, size, rm), strm3)
                 -- CALL rm/X
                 case '010' =>
                    Zfull_inst (p, Zcall (Zrm (rm)), strm3)
                 -- JMP rm/X
                 case '100' =>
                    Zfull_inst (p, Zjmp (rm), strm3)
                 -- PUSH rm/X
                 case '110' =>
                    Zfull_inst (p, Zpush (Zrm (rm)), strm3)
                 case _ => Zdec_fail ("Unsupported opcode: INC/DEC Group 5")
              }
            }

         -- Three byte opcodes
         case 0x0F @ 0x38 @ opc @ strm2 =>
            match opc
            {
               case 0x29 =>
               {
                  if p == list {0x66} then
                  {
                    reg, rm, strm3 = readOpcodeModRM (REX, strm2);
                    a = (reg, rm_to_xmm_mem (rm));
                    Zfull_inst (p, SSE (PCMPEQQ (a)), strm3)
                  }
                  else
                    Zdec_fail ("Unsupported opcode: 0F 38 " : [opc])
               }
               case _ => Zdec_fail ("Unsupported opcode: 0F 38 " : [opc])
            }
         case 0x0F @ 0x3A @ opc @ _ =>
            Zdec_fail ("Unsupported opcode: 0F 3A " : [opc])

         -- Two byte opcodes
         case 0x0F @ opc @ strm2 =>
            match opc
            {
               -- MOVUPS, MOVUPD
               case 0x10 =>
               {
                  reg, rm, strm3 = readOpcodeModRM (REX, strm2);
                  a = (xmm_reg (reg), rm_to_xmm_mem (rm));
                  match p
                  {
                    case list {0x66} =>
                      Zfull_inst (p, SSE (MOVUP_D_S (true, a)), strm3)
                    case list {0xF2} =>
                      Zfull_inst (p, SSE (MOVSD (a)), strm3)
                    case list {0xF3} =>
                      Zfull_inst (p, SSE (MOVSS (a)), strm3)
                    case _ =>
                      Zfull_inst (p, SSE (MOVUP_D_S (false, a)), strm3)
                  }
               }
               case 0x11 =>
               {
                  reg, rm, strm3 = readOpcodeModRM (REX, strm2);
                  a = (rm_to_xmm_mem (rm), xmm_reg (reg));
                  match p
                  {
                    case list {0x66} =>
                      Zfull_inst (p, SSE (MOVUP_D_S (true, a)), strm3)
                    case list {0xF2} =>
                      Zfull_inst (p, SSE (MOVSD (a)), strm3)
                    case list {0xF3} =>
                      Zfull_inst (p, SSE (MOVSS (a)), strm3)
                    case _ =>
                      Zfull_inst (p, SSE (MOVUP_D_S (false, a)), strm3)
                  }
               }

               -- NOP
               case 0x1F =>
               { _, _, strm3 = readOpcodeModRM (REX, strm2);
                 Zfull_inst (p, Znop (Length (strm)), strm3)
               }

               -- MOVAPS, MOVAPD
               case 0x28 =>
               {
                  reg, rm, strm3 = readOpcodeModRM (REX, strm2);
                  double = p == list {0x66};
                  Zfull_inst
                    (p, SSE (MOVAP_D_S (double, xmm_reg (reg),
                                        rm_to_xmm_mem (rm))), strm3)
               }
               case 0x29 =>
               {
                  reg, rm, strm3 = readOpcodeModRM (REX, strm2);
                  double = p == list {0x66};
                  Zfull_inst
                    (p, SSE (MOVAP_D_S (double, rm_to_xmm_mem (rm),
                                        xmm_reg (reg))), strm3)
               }

               -- CVTPI2PS, CVTPI2PD, CVTSI2SS, CVTSI2SD
               case 0x2A =>
               {  reg, rm, strm3 = readOpcodeModRM (REX, strm2);
                  a = (REX.W, reg, rm);
                  match p
                  {
                    case list {0xF3} =>
                      Zfull_inst (p, SSE (CVTSI2SS (a)), strm3)
                    case list {0xF2} =>
                      Zfull_inst (p, SSE (CVTSI2SD (a)), strm3)
                    case _ => Zdec_fail ("Unsupported opcode: 0F " : [opc])
                  }
               }

               -- CVTPS2PI, CVTPD2PI, CVTSS2SI, CVTSD2SI
               case '0x2 110 b' =>
               {  reg, rm, strm3 = readModRM (REX, strm2);
                  a = (not [b], REX.W, reg, rm_to_xmm_mem (rm));
                  match p
                  {
                    case list {0xF3} =>
                      Zfull_inst (p, SSE (CVTSS2SI (a)), strm3)
                    case list {0xF2} =>
                      Zfull_inst (p, SSE (CVTSD2SI (a)), strm3)
                    case _ => Zdec_fail ("Unsupported opcode: 0F " : [opc])
                  }
               }

               -- COMISS, COMISD
               case 0x2F =>
               {
                  reg, rm, strm3 = readOpcodeModRM (REX, strm2);
                  a = (reg, rm_to_xmm_mem (rm));
                  match p
                  {
                    case list {0x66} => Zfull_inst (p, SSE (COMISD (a)), strm3)
                    case _           => Zfull_inst (p, SSE (COMISS (a)), strm3)
                  }
               }

               -- CMOVcc rX, r/mX
               case '0x4 c' =>
               {  size = OpSize (true, REX.W, 1, op_size_override);
                  reg, rm, strm3 = readModRM (REX, strm2);
                  Zfull_inst (p, Zmov ([c], size, Zr_rm (reg, rm)), strm3)
               }

               -- SQRTPD, SQRTPS, SQRTSD, SQRTSS
               case 0x51 =>
               {
                  reg, rm, strm3 = readOpcodeModRM (REX, strm2);
                  a = (reg, rm_to_xmm_mem (rm));
                  match p
                  {
                    case list {0x66} => Zfull_inst (p, SSE (SQRTPD (a)), strm3)
                    case list {0xF3} => Zfull_inst (p, SSE (SQRTSS (a)), strm3)
                    case list {0xF2} => Zfull_inst (p, SSE (SQRTSD (a)), strm3)
                    case _           => Zfull_inst (p, SSE (SQRTPS (a)), strm3)
                  }
               }

               -- ANDPD, ANDPS, ANDNPD, ANDNPS, ORPD, ORPS, XORPD, XORPS
               case '0x5 01 c' =>
               {
                  reg, rm, strm3 = readOpcodeModRM (REX, strm2);
                  a = ([c], reg, rm_to_xmm_mem (rm));
                  match p
                  {
                    case list {0x66} => Zfull_inst (p, SSE (logic_PD(a)), strm3)
                    case _           => Zfull_inst (p, SSE (logic_PS(a)), strm3)
                  }
               }

               -- {ADD|SUB|MUL|DIV|MAX|MIN}{PD|PS|SD|SS}
               -- CVTPD2PS, CVTSS2SD, CVTSD2SS, CVTPS2PD, CVTPS2DQ, CVTPS2DQ
               case '0x5 1 c' =>
               {
                  reg, rm, strm3 = readOpcodeModRM (REX, strm2);
                  a = (reg, rm_to_xmm_mem (rm));
                  match c, p
                  {
                    case 0, list {0x66} =>
                      Zfull_inst (p, SSE (bin_PD (sse_add, a)), strm3)
                    case 0, list {0xF3} =>
                      Zfull_inst (p, SSE (bin_SS (sse_add, a)), strm3)
                    case 0, list {0xF2} =>
                      Zfull_inst (p, SSE (bin_SD (sse_add, a)), strm3)
                    case 0, _ =>
                      Zfull_inst (p, SSE (bin_PS (sse_add, a)), strm3)
                    case 1, list {0x66} =>
                      Zfull_inst (p, SSE (bin_PD (sse_mul, a)), strm3)
                    case 1, list {0xF3} =>
                      Zfull_inst (p, SSE (bin_SS (sse_mul, a)), strm3)
                    case 1, list {0xF2} =>
                      Zfull_inst (p, SSE (bin_SD (sse_mul, a)), strm3)
                    case 1, _ =>
                      Zfull_inst (p, SSE (bin_PS (sse_mul, a)), strm3)
                    case 2, list {0x66} =>
                      Zfull_inst (p, SSE (CVTPD2PS (a)), strm3)
                    case 2, list {0xF3} =>
                      Zfull_inst (p, SSE (CVTSS2SD (a)), strm3)
                    case 2, list {0xF2} =>
                      Zfull_inst (p, SSE (CVTSD2SS (a)), strm3)
                    case 2, _ =>
                      Zfull_inst (p, SSE (CVTPS2PD (a)), strm3)
                    case 3, list {0x66} =>
                      Zfull_inst (p, SSE (CVTPS2DQ (false, a)), strm3)
                    case 3, list {0xF3} =>
                      Zfull_inst (p, SSE (CVTPS2DQ (true, a)), strm3)
                    case 3, list {0xF2} =>
                      Zdec_fail ("Unsupported opcode: 0F " : [opc])
                    case 3, _ =>
                      Zfull_inst (p, SSE (CVTDQ2PS (a)), strm3)
                    case 4, list {0x66} =>
                      Zfull_inst (p, SSE (bin_PD (sse_sub, a)), strm3)
                    case 4, list {0xF3} =>
                      Zfull_inst (p, SSE (bin_SS (sse_sub, a)), strm3)
                    case 4, list {0xF2} =>
                      Zfull_inst (p, SSE (bin_SD (sse_sub, a)), strm3)
                    case 4, _ =>
                      Zfull_inst (p, SSE (bin_PS (sse_sub, a)), strm3)
                    case 5, list {0x66} =>
                      Zfull_inst (p, SSE (bin_PD (sse_min, a)), strm3)
                    case 5, list {0xF3} =>
                      Zfull_inst (p, SSE (bin_SS (sse_min, a)), strm3)
                    case 5, list {0xF2} =>
                      Zfull_inst (p, SSE (bin_SD (sse_min, a)), strm3)
                    case 5, _ =>
                      Zfull_inst (p, SSE (bin_PS (sse_min, a)), strm3)
                    case 6, list {0x66} =>
                      Zfull_inst (p, SSE (bin_PD (sse_div, a)), strm3)
                    case 6, list {0xF3} =>
                      Zfull_inst (p, SSE (bin_SS (sse_div, a)), strm3)
                    case 6, list {0xF2} =>
                      Zfull_inst (p, SSE (bin_SD (sse_div, a)), strm3)
                    case 6, _ =>
                      Zfull_inst (p, SSE (bin_PS (sse_div, a)), strm3)
                    case 7, list {0x66} =>
                      Zfull_inst (p, SSE (bin_PD (sse_max, a)), strm3)
                    case 7, list {0xF3} =>
                      Zfull_inst (p, SSE (bin_SS (sse_max, a)), strm3)
                    case 7, list {0xF2} =>
                      Zfull_inst (p, SSE (bin_SD (sse_max, a)), strm3)
                    case 7, _ =>
                      Zfull_inst (p, SSE (bin_PS (sse_max, a)), strm3)
                  }
               }

               -- Group 12
               case 0x71 =>
                  if p == list {0x66} then
                  {
                      opcode, rm, strm3 = readOpcodeModRM (REX, strm2);
                      imm, strm4 = oimmediate8 (strm3);
                      match opcode, rm
                      {
                         case '010', Zr (r) =>
                           Zfull_inst (p, SSE (PSRLW_imm ([r], [imm])), strm4)
                         case '100', Zr (r) =>
                           Zfull_inst (p, SSE (PSRAW_imm ([r], [imm])), strm4)
                         case '110', Zr (r) =>
                           Zfull_inst (p, SSE (PSLLW_imm ([r], [imm])), strm4)
                         case _ =>
                           Zdec_fail ("Unsupported group 12 opcode: 0F 71")
                      }
                  }
                  else
                    Zdec_fail ("Unsupported group 12 opcode: 0F 71")

               -- Group 13
               case 0x72 =>
                  if p == list {0x66} then
                  {
                      opcode, rm, strm3 = readOpcodeModRM (REX, strm2);
                      imm, strm4 = oimmediate8 (strm3);
                      match opcode, rm
                      {
                         case '010', Zr (r) =>
                           Zfull_inst (p, SSE (PSRLD_imm ([r], [imm])), strm4)
                         case '100', Zr (r) =>
                           Zfull_inst (p, SSE (PSRAD_imm ([r], [imm])), strm4)
                         case '110', Zr (r) =>
                           Zfull_inst (p, SSE (PSLLD_imm ([r], [imm])), strm4)
                         case _ =>
                           Zdec_fail ("Unsupported group 13 opcode: 0F 72")
                      }
                  }
                  else
                    Zdec_fail ("Unsupported group 13 opcode: 0F 71")

               -- Group 14
               case 0x73 =>
                  if p == list {0x66} then
                  {
                      opcode, rm, strm3 = readOpcodeModRM (REX, strm2);
                      imm, strm4 = oimmediate8 (strm3);
                      match opcode, rm
                      {
                         case '010', Zr (r) =>
                           Zfull_inst (p, SSE (PSRLQ_imm ([r], [imm])), strm4)
                         case '011', Zr (r) =>
                           Zfull_inst (p, SSE (PSRLDQ ([r], [imm])), strm4)
                         case '110', Zr (r) =>
                           Zfull_inst (p, SSE (PSLLQ_imm ([r], [imm])), strm4)
                         case '111', Zr (r) =>
                           Zfull_inst (p, SSE (PSLLDQ ([r], [imm])), strm4)
                         case _ =>
                           Zdec_fail ("Unsupported group 14 opcode: 0F 73")
                      }
                  }
                  else
                    Zdec_fail ("Unsupported group 14 opcode: 0F 71")

               -- MOVQ
               case 0xD6 =>
               {
                  reg, rm, strm3 = readOpcodeModRM (REX, strm2);
                  a = (rm_to_xmm_mem (rm), xmm_reg ([reg]));
                  match p
                  {
                    case list {0x66} => Zfull_inst (p, SSE (MOVQ (a)), strm3)
                    case _ => Zdec_fail ("Unsupported opcode: 0F " : [opc])
                  }
               }

               -- MOVD, MOVQ
               case '011 b 0xE' =>
               {
                  reg, rm, strm3 = readOpcodeModRM (REX, strm2);
                  a = (not [b], REX.W, reg, rm);
                  match p
                  {
                    case list {0x66} => Zfull_inst (p, SSE (MOV_D_Q (a)), strm3)
                    case list {0xF3} =>
                      if b == '1' then
                        Zfull_inst
                          (p, SSE (MOVQ (xmm_reg ([reg]), rm_to_xmm_mem (rm))),
                           strm3)
                      else
                        Zdec_fail ("Unsupported opcode: 0F " : [opc])
                    case _ => Zdec_fail ("Unsupported opcode: 0F " : [opc])
                  }
               }

               -- Jcc rel32
               case '0x8 c' =>
               {  imm, strm3 = immediate32 (strm2);
                  Zfull_inst (p, Zjcc ([c], imm), strm3)
               }

               -- SETcc r/m8
               case '0x9 c' =>
               {
                  _, rm, strm3 = readModRM (REX, strm2);
                  Zfull_inst (p, Zset ([c], have_rex, rm), strm3)
               }

               -- BT{S,R,C} r/mX, r
               case '101 x 011' => -- A3, AB, B3, BB
               {  size = OpSize (have_rex, REX.W, 1, op_size_override);
                  reg, rm, strm3 = readModRM (REX, strm2);
                  Zfull_inst (p, Zbit_test ([x], size, Zrm_r (rm, reg)), strm3)
               }

               case 0xBA =>
               {  size = OpSize (have_rex, REX.W, 1, op_size_override);
                  opcode, rm, strm3 = readOpcodeModRM (REX, strm2);
                  imm, strm4 = oimmediate8 (strm3);
                  match opcode
                  {  -- BT{S,R,C} r/mX, imm8
                     case '1 x' =>
                        Zfull_inst
                          (p, Zbit_test ([x], size, Zrm_i (rm, imm)), strm4)
                     case _ => Zdec_fail ("Unsupported opcode: Group 8")
                  }
               }

               -- IMUL rX, r/mX
               case 0xAF =>
               {  size = OpSize (have_rex, REX.W, 1, op_size_override);
                  reg, rm, strm3 = readModRM (REX, strm2);
                  Zfull_inst (p, Zimul2 (size, reg, rm), strm3)
               }

               -- CMPXCHG r/mX, rX
               case '0xB 000 v' =>
               {  size = OpSize (have_rex, REX.W, v, op_size_override);
                  reg, rm, strm3 = readModRM (REX, strm2);
                  Zfull_inst (p, Zcmpxchg (size, rm, reg), strm3)
               }

               -- MOVZX rX, r/m8
               -- MOVZX rX, r/m16
               -- MOVSX rX, r/m8
               -- MOVSX rX, r/m16
               case '0xB s 11 v' =>
               {  size2 = OpSize (have_rex, REX.W, 1, op_size_override);
                  size = if v == 1 then Z16 else Z8 (have_rex);
                  reg, rm, strm3 = readModRM (REX, strm2);
                  arg = (size, Zr_rm (reg, rm), size2);
                  instr = if s == '1' then Zmovsx (arg) else Zmovzx (arg);
                  Zfull_inst (p, instr, strm3)
               }

               -- XADD r/mX, rX
               case '0xC 000 v' =>
               {  size = OpSize (have_rex, REX.W, v, op_size_override);
                  reg, rm, strm3 = readModRM (REX, strm2);
                  Zfull_inst (p, Zxadd (size, rm, reg), strm3)
               }

               -- CMPPS, CMPPD, CMPSS, CMPSD
               case 0xC2 =>
               {
                  reg, rm, strm3 = readOpcodeModRM (REX, strm2);
                  imm, strm4 = oimmediate8 (strm3);
                  a = ([imm], reg, rm_to_xmm_mem (rm));
                  match p
                  {
                    case list {0x66} => Zfull_inst (p, SSE (CMPPD (a)), strm4)
                    case list {0xF3} => Zfull_inst (p, SSE (CMPSS (a)), strm4)
                    case list {0xF2} => Zfull_inst (p, SSE (CMPSD (a)), strm4)
                    case _           => Zfull_inst (p, SSE (CMPPS (a)), strm4)
                  }
               }

               -- CVTPD2DQ, CVTDQ2PD, CVTTPD2DQ
               case 0xE6 =>
               {
                  reg, rm, strm3 = readOpcodeModRM (REX, strm2);
                  a = (reg, rm_to_xmm_mem (rm));
                  match p
                  {
                    case list {0x66} =>
                      Zfull_inst (p, SSE (CVTPD2DQ (true, a)), strm3)
                    case list {0xF3} =>
                      Zfull_inst (p, SSE (CVTDQ2PD (a)), strm3)
                    case list {0xF2} =>
                      Zfull_inst (p, SSE (CVTPD2DQ (false, a)), strm3)
                    case _ => Zdec_fail ("Unsupported opcode: " : [opc])
                  }
               }

              case _ => Zdec_fail ("Unsupported opcode: 0F " : [opc])
            }

         -- Unsupported
         case opc @ _ =>
            Zdec_fail ("Unsupported opcode: " : [opc])

         case Nil => Zdec_fail ("No opcode")
       }
  }
}

---------------------------------------------------------------------------

-- ===================
-- Next State Function
-- ===================

stream x64_fetch =
{
   var strm = Nil;
   for i in 19 .. 0 do strm <- Cons (MEM (RIP + [i]), strm);
   strm
}

unit x64_next =
   match x64_decode (x64_fetch)
   {
      case Zfull_inst (_, i, Some (strm1)) =>
        {
           len = 20 - Length (strm1);
           RIP <- RIP + [len];
           Run (i)
        }
      case Zfull_inst (_, _, None) => #FAILURE "not enough bytes"
      case Zdec_fail (s) => #FAILURE (s)
   }

---------------------------------------------------------------------------
